/********************************************************************************
Rockland
1999-oo Tito Dal Canton
	Calcoli geometrici vari utili ovunque e sempre in R^2 e R^3.
********************************************************************************/
#include <stdio.h>
#include <math.h>
#include "global.h"
#include "3d-calc.h"

// dividi vettore 3d per la sua norma
void Normalize(Vector3D *v)
{
	float	value = sqrt(v->x * v->x + v->y * v->y + v->z * v->z);
	
	if (value != 0) {
		float	inv = 1.0 / value;
		
		v->x *= inv;
		v->y *= inv;
		v->z *= inv;
	}
}

// dividi vettore 2d per la sua norma
void Normalize2D(Vector2D *v)
{
	float	value = sqrt(v->x * v->x + v->y * v->y);
	
	if (value != 0) {
		float	inv = 1.0 / value;

		v->x *= inv;
		v->y *= inv;
	}
}

// prodotto vettoriale
void CrossProduct(Vector3D a, Vector3D b, Vector3D *out)
{
	out->x = (a.y * b.z) - (a.z * b.y);
	out->y = (a.z * b.x) - (a.x * b.z);
	out->z = (a.x * b.y) - (a.y * b.x);
}

// modulo o norma del vettore
float VectorModule(Vector3D v)
{
	return sqrt(v.x * v.x + v.y * v.y + v.z * v.z);
}

// prodotto per scalare
void VectorMult(Vector3D *v, float value)
{
	v->x *= value;
	v->y *= value;
	v->z *= value;
}

// calcola la proiezione di un vettore <v> su un piano _|_ a <pn>
// lungo la direzione <pd>. Per utilita' varia restituisce anche alfa
// (il parametro tale che P appartiene alla retta <v> e al piano).
float ProjectVector(Vector3D *v, Vector3D *pd, Vector3D *pn, Vector3D *out)
{
	float	alfa, div;

	div = DotProduct(*pd, *pn);
	if (div == 0) {	// direzione || al piano -> proiezione infinita!!
		out->x = out->y = out->z = 0;
		return 0;
	} else {
		alfa = -DotProduct(*v, *pn) / div;
		out->x = v->x + alfa * pd->x;
		out->y = v->y + alfa * pd->y;
		out->z = v->z + alfa * pd->z;
		return alfa;
	}
}

// controlla se un punto nello spazio "tocca" un poligono,
// in pratica se il punto cade all'interno del prisma che ha come
// base il poligono stesso. Usata generalmente per capire se
// una proiezione cade dentro o fuori al poligono su cui e' stata fatta.
int PointTouchesPoly(Point3D *p, Polygon3D *poly)
{
	Vector3D		n = poly->normal;
	unsigned int	i;

#if POLYGONS_CCW_IN_RML
	VectorMult(&n, -1);
#endif
	// controlla se il punto e' "dietro" a qualche lato del poligono
	for (i = 0; i < poly->numVertices; i++) {
		Vector3D	ab, faceNorm, ap;
		Point3D		a;

		a.x = poly->vertices[i].x;
		a.y = poly->vertices[i].y;
		a.z = poly->vertices[i].z;
		// calcola il vettore normale al lato in esame
		if (i == poly->numVertices - 1) {
			ab.x = poly->vertices[0].x - a.x;
			ab.y = poly->vertices[0].y - a.y;
			ab.z = poly->vertices[0].z - a.z;
		} else {
			ab.x = poly->vertices[i+1].x - a.x;
			ab.y = poly->vertices[i+1].y - a.y;
			ab.z = poly->vertices[i+1].z - a.z;
		}
		ap.x = p->x - a.x;
		ap.y = p->y - a.y;
		ap.z = p->z - a.z;
		CrossProduct(n, ab, &faceNorm);
		Normalize(&faceNorm);
		if (DotProduct(ap, faceNorm) < 0.0)
			return 0;
	}
	return 1;
}

