#define	NONE				-1
#define	FLOOR				-2
#define	CEILING				-3
#define	SMALL_DIFF			1.0e-10
#define POLYGONS_CCW_IN_RML	1

//----
typedef struct {	// vertice di un poligono con texture
	float	x, y, z;	// posizione
	float	u, v;		// coordinate texture
	float	lm_u, lm_v;	// coordinate lightmap
	float	nx, ny, nz;	// normale
} Vertex3D;

typedef struct {	// Punto 3D e vettore 3D
	float	x;
	float	y;
	float	z;
} Point3D, Vector3D;

typedef struct {	// Punto e vettore 2D
	float	x;
	float	y;
} Point2D, Vector2D;

//----
#define kMaxVertices 		8

typedef struct {	// poligono per l'effetto-ombra
	unsigned int	numVertices;
	Point3D			vertices[kMaxVertices];
} ShadowPoly;

typedef struct {	// poligono con texture
	unsigned int	numVertices;
	int				texture;
	int				id;
	float			r, g, b, a;
	Vertex3D		vertices[kMaxVertices];
	
	Vector3D		normal;	// _|_ al piano su cui giace il poligono
	float			dist;

	unsigned int	lmap;	// dovrebbe essere GLuint
} Polygon3D;

//----
typedef struct Camera {	// punto di osservazione (posizione eccetera)
	float	x;
	float	y;
	float	z;
	float	radius;
	float	height;
	float	ySpeed;
	float	floorY;
	float	angle;
	float	elevation;
} Camera;

//----
typedef struct MapSection {		// Pavimento + soffitto = sezione minima della mappa
	long		numVertices;
	Point2D		vertices[kMaxVertices];
	// per vedere se un punto e' nella map section:
	Vector2D	wallsNormals[kMaxVertices];	// normali delle pareti
	float		wallsDists[kMaxVertices];	// e distanze
	unsigned long	connectedSides;	// mappa di bit: ognuno indica se il corrispondente lato
									// della section e' connesso con un'altra section. TODO...
	float		floorY;
	float		ceilingY;
	// liquido
	float		liquidHeight;
	// utilities
	int			floorPolyID;	// indice del pavimento
	int			ceilingPolyID;	// indice del soffitto
	// suono ambiente
	int			ambientSound;
	long		ambientVolume;
} MapSection;

//----
typedef struct {
	float		diffuse[4];
	float		ambient[4];
	Vector3D	dir;
	Point3D		pos;
} LightSource;

typedef struct {
	int		on;
	float	dist;
	float	color[3];
} FogParams;

//----

// gDisplayFlags
enum {
	SHADOWS = 1,
	SHOW_NORMALS = 2,
	SHOW_INFO = 4
};

