/********************************************************************************
Rockland
1998-oo Tito Dal Canton
	Motore 3D per esperimenti di creazione mondi
********************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <SDL.h>
#include <SDL_opengl.h>
#include "global.h"
#include "map.h"
#include "render.h"
#include "worlddata.h"
#include "rocklandinput.h"
#include "main-menu.h"
#include "textures.h"
#include "font-engine.h"
#include "3d-calc.h"
#include "objects.h"

static void RocklandRun(void);
static void help(void);

SDL_Surface			*main_surface = NULL;
int					gDone;
extern Polygon3D	gStaticPolys[];
extern long			gStaticPolysCount;
extern MapSection	gMap[kMaxMapSections];
extern int			gMapSectionsNum;
int					gScreenWidth = 800, gScreenHeight = 600;
extern Camera		gCamera;
int					currMapSection = NONE;
unsigned long		gDisplayFlags = SHOW_INFO;
extern Object		*gFirstObject;
float				gCameraSpeedAcc = 0;
GLuint				textures[kNumTextures];
extern GLfloat		gTeleportEffectDecay;
float				fps = 0.0;

int main(int argc, char *argv[])
{
	int	toDo = 0, err = 0, i;
	int	mapNameIndex = -1;
	int	fullscreen = 0;

	fprintf(stderr, "Rockland - Motore 3D per creare mondi sperimentali\n1998-oo Tito Dal Canton - www.dalcanton.it/tito\n\n");
	srandom(time(NULL));
	
	// parsing dei parametri da riga di comando
	for (i = 1; i < argc; i++) {
		if (!strcmp(argv[i], "-m") || !strcmp(argv[1], "--map")) {
			if (argc > i+1) {
				i++;
				mapNameIndex = i;
			}
		} else if (!strcmp(argv[i], "-f")) {
			fullscreen = 1;
		} else if (!strcmp(argv[i], "-n") || !strcmp(argv[i], "--normals")) {
			gDisplayFlags |= SHOW_NORMALS;
		} else if (!strcmp(argv[i], "-d") || !strcmp(argv[i], "--shadows")) {
			gDisplayFlags |= SHADOWS;
		} else if (!strcmp(argv[i], "-s")) {
			if (argc > i+1) {
				int     j = 0, width, height;
				char    *str, first_digit[10] = "";
				
				i++;
				str = argv[i];
				while (str[j] != 'x')
					j++;
				strncpy(first_digit, str, j);
				width = atoi(first_digit);
				height = atoi(str+j+1);
				gScreenWidth = width;
				gScreenHeight = height;
				fprintf(stderr, "Risoluzione impostata a %dx%d\n", width, height);
			}
		} else if (!strcmp(argv[i], "-h") || !strcmp(argv[1], "--help")) {
			help();
			return 0;
		}
	}
	
	// inizializza le SDL
	fprintf(stderr, "Inizializzo SDL...");
	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		fprintf(stderr, " Errore: %s\n", SDL_GetError());
		return 1;
	}
	fprintf(stderr, "\n");
	SDL_WM_SetCaption("Rockland", NULL);
	// setta i parametri di OpenGL:
	// 5 bit per colore, 16 bit per la profondita', double buffering
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	// imposta la modalita' video che ci serve
	main_surface = SDL_SetVideoMode(gScreenWidth, gScreenHeight, 0, SDL_OPENGLBLIT | (fullscreen ? SDL_FULLSCREEN : 0));
	if (main_surface == NULL) {
		fprintf(stderr, "Non posso ottenere la risoluzione %dx%d con OpenGL: %s\n", gScreenWidth, gScreenHeight, SDL_GetError());
		return 1;
	}
	SDL_EnableKeyRepeat(1, 0);
	SDL_EventState(SDL_MOUSEMOTION, SDL_IGNORE);

	// inizializza il resto
	font_init();
	for (i = 0; i < kNumTextures; i++)
		textures[i] = 0;
	InitObjects();
	
	gDone = 0;
	fprintf(stderr, "\n");
	do {
		DrawMainMenu(1);
		toDo = DoMainMenu();
		if (toDo == 1) {
			// via!
			fprintf(stderr, "Inizio la simulazione.\n");
			SDL_WarpMouse(gScreenWidth/2, gScreenHeight/2);
			err = LoadMapFile(mapNameIndex != -1 ? argv[mapNameIndex] : "original-complex.rml");
			InitOpenGL();
			if (err)
				break;
			if (fullscreen)	SDL_ShowCursor(SDL_DISABLE);
			RocklandRun();
			if (fullscreen)	SDL_ShowCursor(SDL_ENABLE);
			QuitOpenGL();
		}
	} while (toDo != 3);
	fprintf(stderr, "\nFine.\n");
	SDL_Quit();
	return 0;
}

static void RocklandRun(void)
{
	int					tenFrames = 0;
	unsigned long		prevTime, currTime;
	unsigned long		tenStart = 0;
	unsigned long		lastShotTime = 0;
	RocklandEventData	inputData = {0, 0, 0.0, 0.0, 0};

	// trova la map section iniziale
	{
		int	i;
	
		currMapSection = NONE;
		for (i = 0; i < gMapSectionsNum; i++) {
			if (PointInMapSection(&gMap[i], gCamera.x, gCamera.y, gCamera.z)) {
				currMapSection = i;
				break;
			}
		}
		if (currMapSection == NONE) {
			fprintf(stderr, "Il punto di partenza e' al di fuori delle map sections!\n"
							"Bisogna correggere la descrizione del mondo.\n");
			return;
		} else {
			fprintf(stderr, "Trovata la map section iniziale (%d)\n", i);
		}
	}
	// effetto teletrasporto (esperimenti vari)
	TeleportEffect();
	prevTime = SDL_GetTicks();
	tenStart = prevTime;
	lastShotTime = prevTime;
	gTeleportEffectDecay = 1.0;
	// main l00p
	while (!inputData.quit) {
		float		currX, currY, currZ;
		float		currAngle, currElevation, rad_angle, rad_elev;
		double		dt;
		Vector3D	pos_delta = {0, 0, 0};
		
		// calcola un valore che permetta di regolare le varie velocita'
		// in modo che siano indipendenti dalla velocita' effettiva
		// della macchina. Questo valore lo decidiamo in base al tempo
		// impiegato per processare un fotogramma.
		currTime = SDL_GetTicks();
		dt = (double)(currTime - prevTime);
		prevTime = currTime;
		if (gTeleportEffectDecay > 0.0)
			gTeleportEffectDecay -= 0.001 * dt;
		
		// interroga il sistema di input
		QueryEvents(&inputData);

		currX = gCamera.x;
		currY = gCamera.y;
		currZ = gCamera.z;
		currAngle = gCamera.angle;
		currElevation = gCamera.elevation;
		rad_angle = currAngle * (M_PI / 180.0);
		rad_elev = currElevation * (M_PI / 180.0);
		// controlla i comandi avanti-indietro
		if (inputData.fwd) {
			if (gCameraSpeedAcc < kCameraMaxSpeed)
				gCameraSpeedAcc += 0.003 * dt;
			if (gCameraSpeedAcc > kCameraMaxSpeed)
				gCameraSpeedAcc = kCameraMaxSpeed;
			pos_delta.x = sin(rad_angle) * gCameraSpeedAcc * dt;
			pos_delta.z = cos(rad_angle) * gCameraSpeedAcc * dt;
		} else if (inputData.bwd) {
			if (gCameraSpeedAcc > -kCameraMaxSpeed)
				gCameraSpeedAcc -= 0.003 * dt;
			if (gCameraSpeedAcc < -kCameraMaxSpeed)
				gCameraSpeedAcc = -kCameraMaxSpeed;
			pos_delta.x = sin(rad_angle) * gCameraSpeedAcc * dt;
			pos_delta.z = cos(rad_angle) * gCameraSpeedAcc * dt;
		} else {
			// decelera progressivamente fino a velocita' zero
			if (fabsf(gCameraSpeedAcc) > 0.0) {
				gCameraSpeedAcc *= 0.75;
				pos_delta.x = sin(rad_angle) * gCameraSpeedAcc * dt;
				pos_delta.z = cos(rad_angle) * gCameraSpeedAcc * dt;
			}
		}
		// ed ora quelli per girarsi (destra/sinistra - su/giu')
		currAngle += inputData.rotationDelta * dt;
		if (currAngle > 360.0)
			currAngle = currAngle - 360.0;
		else if (currAngle < 0)
			currAngle = 360.0 + currAngle;
		currElevation += inputData.elevationDelta * dt;
		if (currElevation < -60.0)
			currElevation = -60.0;
		if (currElevation > 60.0)
			currElevation = 60.0;
		// leggera variazione dell'altezza per dare l'effetto del respiro/passo
		currY += sin(currTime / 300.0) * 0.002;

		// +- TODO (collision detect) --------------------------------------+
		// | anziche' rompersi la testa con n collisioni possibili, sarebbe |
		// | bello calcolare in un primo tempo tutte le forze che agiscono  |
		// | sul giocatore, e variare l'accelerazione del giocatore stesso  |
		// | in base alla forza risultante e ad una sua ipotetica massa...  |
		// | esattamente come accade nel mondo reale :)                     |
		// | A questo punto potremmo addirittura applicare direttamente     |
		// | anche la forza di attrito (il cui coefficiente magari varia    |
		// | per ogni texture del pavimento ;-) e risolviamo n problemi in  |
		// | un colpo solo.
		// +----------------------------------------------------------------+
#if 1
		// controlla se siamo sospesi in aria, se si applica la gravita'
		// Sistema ERRONEO e poco elegante.
		{
			float	floorY = gMap[currMapSection].floorY;
			float	camHeight = gCamera.height;
			float	deltaY, newY;
			
			deltaY = (currY - camHeight) - floorY;
			if (deltaY > 0)	{	// stiamo volando, facci cadere
				newY = currY - kClimbSpeed * dt;
				if (((newY - camHeight) - floorY) < 0)
					newY = floorY + camHeight;
			} else if (deltaY < 0) {	// siamo immersi nel pavimento :/
				newY = currY + kClimbSpeed * dt;
				if (((newY - camHeight) - floorY) > 0)
					newY = floorY + camHeight;
			}
			pos_delta.y = newY - currY;
		}
#else
		// questo sarebbe piu' corretto dal punto di vista fisico:
		// infatti nel mondo reale veniamo continuamente 'tirati giu'
		// dalla gravita', e collidiamo col terreno
		pos_delta.y = -0.0001 * dt;
#endif

		//---- c o l l i s i o n  d e t e c t ----
		{
			int		escapingWorld = 0, hitSide = NONE;
			Point3D	new_pos;

			new_pos.x = currX + pos_delta.x;
			new_pos.y = currY + pos_delta.y;
			new_pos.z = currZ + pos_delta.z;
			// collisioni con qualche oggetto?
			CheckPlayerObjectCollision(new_pos, &pos_delta);
			
			// gestisci le collisioni giocatore-mondo
			if (!SphereInMapSection(currMapSection, new_pos.x, new_pos.y, new_pos.z, gCamera.radius, &hitSide)) {
				// la sfera circoscritta alla camera non e' piu' nella section in cui e' stata finora
				escapingWorld = 1;
				if (hitSide > -1) {
					Vector2D	normal = gMap[currMapSection].wallsNormals[hitSide];
					Vector2D	direction;
					float		dot;

					// controlla se ci stiamo allontanando dall'ostacolo:
					// in tal caso non bloccare la camera
					direction.x = pos_delta.x;
					direction.y = pos_delta.z;
					dot = (normal.x * direction.x) + (normal.y * direction.y);
					if (dot < 0)
						escapingWorld = 0;
				}
				if (!PointInMapSection(&gMap[currMapSection], new_pos.x, new_pos.y, new_pos.z)) {
					// addirittura il centro della camera e' fuori dalla section
					int	i, newMapSection = NONE;

					// cerca se la camera e' finita in un'altra section
					for (i = 0; i < gMapSectionsNum; i++) {
						if (i != currMapSection && PointInMapSection(&gMap[i], new_pos.x, new_pos.y, new_pos.z)) {
							newMapSection = i;
							break;
						}
					}
					if (newMapSection >= 0) {
						// trovata la nuova section, controlla se ci stiamo in altezza
						if ((gMap[newMapSection].ceilingY - gMap[newMapSection].floorY) <= gCamera.height) {
							SlideInsideMapSection(currMapSection, currX, currY, currZ, &pos_delta);
						} else if ((gMap[newMapSection].floorY - gMap[currMapSection].floorY) > kMaxClimbHeight) {
							// il pavimento e' troppo alto, non riusciamo a salire
							SlideInsideMapSection(currMapSection, currX, currY, currZ, &pos_delta);
						} else {
							currMapSection = newMapSection;
						}
					} else {
						// la camera e' completamente uscita dal mondo
						SlideInsideMapSection(currMapSection, currX, currY, currZ, &pos_delta);
					}
					escapingWorld = 0;
				} else if (escapingWorld && (hitSide != NONE)) {
					// il centro della camera e' nella vecchia section, ma la sfera circoscritta
					// ne ha attraversato un lato. Bisogna capire se questo lato e' attraversabile
					// (porta ad un'altra section) oppure se ci dobbiamo fermare.
					Vector2D	v;
					float		test_x, test_y, test_z;
					int			i;

					if (hitSide == FLOOR || hitSide == CEILING) {
						pos_delta.y = 0;
					} else {
					// calcola un punto della sfera sicuramente
					// fuori dalla section corrente
					v = gMap[currMapSection].wallsNormals[hitSide];
					v.x *= gCamera.radius;
					v.y *= gCamera.radius;
					test_x = v.x + currX;
					test_y = currY;
					test_z = v.y + currZ;
					// ...e controlla se questo punto appartiene ad un'altra section
					// (ebbene si, e' un metodo ridicolo ma funziona :)
					for (i = 0; i < gMapSectionsNum; i++) {
						if (PointInMapSection(&gMap[i], test_x, test_y, test_z)) {
							// si, appartiene alla section i.
							if (i != currMapSection) {
								if ((gMap[i].floorY - gMap[currMapSection].floorY) > kMaxClimbHeight) {
									// il pavimento e' troppo alto, non possiamo arrampicarci
									SlideInsideMapSection2(currMapSection, currX, currY, currZ, hitSide, &pos_delta);
								}
								// FIXME: considerare il caso del soffitto troppo basso
							}
							escapingWorld = 0;
							break;
						}
					}
					if (escapingWorld) {
						// abbiamo sbattuto contro una parete, scivola...
						SlideInsideMapSection2(currMapSection, currX, currY, currZ, hitSide, &pos_delta);
						escapingWorld = 0;
					}
					}
				}
			}
		}

		// applica i movimenti alla nostra posizione corrente
		gCamera.x += pos_delta.x;
		gCamera.y += pos_delta.y;
		gCamera.z += pos_delta.z;
		gCamera.angle = currAngle;
		gCamera.elevation = currElevation;

		// controlla lo sparo
		if (inputData.fire && (SDL_GetTicks() - lastShotTime) > 500) {
			Point3D     pos = {currX, (currY - 0.2), currZ};
			Vector3D    speed;
			
			speed.x = sin(rad_angle) * 0.6;
			speed.y = sin(-rad_elev) * 0.6;
			speed.z = cos(rad_angle) * 0.6;
			NewObject(MODEL_FIREBALL, pos, speed, 0, 0, 0);	// sarebbero da scegliere valori casuali per gli angoli
			lastShotTime = SDL_GetTicks();
		}
		// controlla il tasto azione e invia
		// l'eventuale azione agli oggetti
		if (inputData.action)
			BroadcastObjectAction();
		
		UpdateObjects(dt);

		// ogni 10 frames calcola un nuovo valore di fps
		tenFrames++;
		if (tenFrames == 10) {
			tenFrames = 0;
			fps = currTime - tenStart;  // millisecondi per disegnare 10 frames
			fps *= 0.1;                 // millisecondi per disegnare 1 frame
			fps = 1000.0 / fps;         // frames al secondo
			tenStart = currTime;
		}

		// renderizza tutto quanto
		RocklandRender();
	}
}

static void help(void)
{
	printf(	"Sintassi:\n\n"
			"    rockland [opzioni]\n\n"
			"Opzioni riconosciute:\n\n"
			"    -m mappa.rml Visualizza il mondo descritto dal file mappa.rml anzich quello originale.\n"
			"    -s WxH       Specifica la risoluzione video (p.es. -s 640x480). La risoluzione di default  800x600\n"
			"    -f           Visualizzazione a pieno schermo (consigliata).\n"
			"    -d           Mostra le ombre generate dalla luce del sole. Ancora imperfetto, ma relativamente bello...\n"
			"    -n           Debug: mostra i vettori normali ai vertici del mondo.\n"
		  );
}

