/*****************************************************************************************************
Rockland - 1998-oo Tito Dal Canton
	Routine di gestione e organizzazione del mondo.
	Il mondo di Rockland e' suddiviso in poligoni e 'map sections'. I poligoni costituiscono
	quello che "si vede" del mondo, mentre le map sections delimitano lo spazio in cui
	si puo' liberamente scorrazzare. Il bello e' che poligoni e ms possono essere completamente
	diversi: e' possibile cioe' fare una stanza rettangolare di poligoni, ma limitare i movimenti
	come se fosse rotonda. Ovviamente cio' non ha una grande utilita', ma un'organizzazione
	del genere e' abbastanza semplice da gestire. Per quello che dobbiamo fare non credo che
	serva implementare un mondo ad albero binario come Quake :-)
	Comunque, la lista di poligoni viene renderizzata in sequenza, senza tante ottimizzazioni
	(tanto OpenGL e' velocissimo, e anche i Mac moderni lo sono). La lista di map sections viene
	invece usata ad ogni spostamento della telecamera: date le nuove coordinate, esse vengono
	confrontate con _ogni_ map section, per capire in quale siamo finiti (NOTA: con il nuovo sistema
	iber-efficiente controlliamo tutta la lista solo quando siamo sicuri di essere usciti dalla
	section in cui eravamo, velocizzando le cose).
	Sapendo questo possiamo regolare l'altezza della camera a seconda dell'altezza del pavimento
	corrispondente a quella map section. Se risulta che le nuove coordinate non sono in nessuna section,
	assumiamo che la camera ha sbattuto il naso contro una parete, e non aggiorniamo la sua posizione.
	Mi sembra un sistema veramente primitivo... probabilmente non saro' mai soddisfatto.
******************************************************************************************************/
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <SDL_opengl.h>
#include "global.h"
#include "3d-calc.h"
#include "map.h"
#include "worlddata.h"
#include "objects.h"
#include "textures.h"

#define CALC_VERTEX_NORMALS	1

static int GetNextTag(FILE *f, char *buf);
static void ClipShadowToPlane(ShadowPoly *in, Vector3D *normal, float dist, ShadowPoly *out);

Polygon3D			gStaticPolys[kNumPolygons];
long				gStaticPolysCount = 0;
MapSection			gMap[kMaxMapSections];
int					gMapSectionsNum = 0;
Camera				gCamera;
extern float		clearColor[];
extern FogParams	gFogParams;
extern LightSource	gGlobalLight;
extern GLuint		textures[];
extern GLuint		gSkyTexture;

/**** PARSER DELLA DESCRIZIONE DEL MONDO (file .rml) ****/

// trova il prossimo tag, saltando i caratteri 'blank' e i commenti, e copia il contenuto del tag in buf
// 0 = tutto ok
// -1 = eof
static int GetNextTag(FILE *f, char *buf)
{
	char	c;
	int		i = 0;
	
	do {
		c = getc(f);
		if (c == '<') {					// inizia un tag
			c = getc(f);				// leggi il primo car
			if (c == '!') {
				// inizia un commento, salta tutti i caratteri
				// fino alla fine del commento (--> come in html)
				while (1) {
					if (getc(f) == '-')				// -.
						if (getc(f) == '-')			//  |
							if (getc(f) == '>')		//  |--> orribile
								break;				//  |
				}									// -'
				continue;
			}
			i = 0;
			buf[i++] = c;				// metti il car nel buffer e inizia a copiare gli altri
			do {
				c = getc(f);
				if (c != '>')
					buf[i++] = c;		// finche' non finisce il tag, copia il contenuto nel buffer di uscita
			} while (c != '>');
			buf[i] = '\0';				// termina la stringa
			return 0;
		} else if (c == EOF) {
			return -1;
		}
	} while (1);
}

// carica il file specificato, inizializza tutte le strutture dati e ricava poligoni e map sections,
// eventualmente facendo qualche pre-calcolo come ombre e normali...
// FIXME:
// - il numero massimo di poligoni e' statico, eliminare questa grave mancanza
//   implementando una allocazione dinamica della memoria necessaria.
//   Penso che una lista linkata vada ugualmente bene.
// - il campo <id> del tag 'poly' non viene considerato, cioe' il poligono
//   viene comunque messo in coda alla lista corrente. Questo non va troppo bene...
// - IMPLEMENTARE UN BSP TREE :D
int LoadMapFile(char *name)
{
	FILE	*f;
	char	tagBuf[50] = "";
	int		i;
	int		hireslmcount = 0;
	
	gCamera.x = 0;	gCamera.y = 0;	gCamera.z = 0;
	gCamera.angle = 0;	gCamera.elevation = 0;
	gCamera.radius = 0.1;
	gStaticPolysCount = 0;
	gMapSectionsNum = 0;
	gFogParams.on = 0;
	
	// apri il file
	if ((f = fopen(name, "rt")) == NULL)
		return -1;
	
	GetNextTag(f, tagBuf);
	if (strcmp(tagBuf, "rockland-map") != 0) {	// il primo tag deve essere <rockland-map>
		fclose(f);
		fprintf(stderr, "%s non sembra la descrizione di un mondo Rockland\n", name);
		return -1;
	}
	do {
		int	err = 0;
	
		err = GetNextTag(f, tagBuf);
		if (err != 0 || strcmp(tagBuf, "/rockland-map") == 0)	// fine mappa
			break;
		if (strcmp(tagBuf, "start") == 0) {						// start
			float	camHeight = 0.0;
			
			do {
				err = GetNextTag(f, tagBuf);
				if (strcmp(tagBuf, "/start") == 0 || err)
					break;
				if (strncmp(tagBuf, "position ", 9) == 0) {
					sscanf(tagBuf+9, "%f %f %f", &gCamera.x, &gCamera.y, &gCamera.z);
				}
				if (strncmp(tagBuf, "elevation ", 10) == 0)
					sscanf(tagBuf+10, "%f", &gCamera.elevation);
				if (strncmp(tagBuf, "azimuth ", 8) == 0) {
					sscanf(tagBuf+8, "%f", &gCamera.angle);
					gCamera.angle += 270;
				}
				if (strncmp(tagBuf, "cam-height ", 11) == 0)
					sscanf(tagBuf+11, "%f", &camHeight);
			} while (1);
			gCamera.y += camHeight;
			gCamera.height = camHeight;
		} else if (strcmp(tagBuf, "polygons") == 0) {			// sezione poligoni
			err = GetNextTag(f, tagBuf);
			while (err == 0 && strcmp(tagBuf, "/polygons") != 0) {
				Polygon3D	poly;
				
				if (strncmp(tagBuf, "poly", 4) == 0) {			// poly
					int	v = 0;
					
					poly.r = poly.g = poly.b = poly.a = 1.0;
					poly.id = NONE;
					do {
						err = GetNextTag(f, tagBuf);
						if (strncmp(tagBuf, "vert ", 5) == 0) {	// vert x y z u v
							sscanf(tagBuf+5, "%f %f %f %f %f", &poly.vertices[v].x, &poly.vertices[v].y, &poly.vertices[v].z,
																&poly.vertices[v].u, &poly.vertices[v].v);
							v++;
						} else if (strncmp(tagBuf, "text ", 5) == 0) {	// texture id
							sscanf(tagBuf+5, "%d", &poly.texture);
							if (poly.texture != NONE && textures[poly.texture] == 0) {
								// texture non ancora caricata, fallo ora
								char	filename[1024] = "";

								sprintf(filename, "textures/tex%d", poly.texture);
								fprintf(stderr, "Richiesta texture \"%s\"... ", filename);
								textures[poly.texture] = LoadTexture(filename, ALPHAMODE_OPAQUE, GL_LINEAR);
								fprintf(stderr, "caricata\n");
							}
						} else if (strncmp(tagBuf, "id ", 3) == 0) {
							sscanf(tagBuf+3, "%d", &poly.id);
						} else if (strncmp(tagBuf, "rgba ", 4) == 0) {
							sscanf(tagBuf+4, "%f %f %f %f", &poly.r, &poly.g, &poly.b, &poly.a);
						}
					} while (err == 0 && strcmp(tagBuf, "/poly") != 0);
					if (v > 2) {
						float       p1x, p1y, p1z;
						Vector3D	n, pq, pr;
						
						poly.numVertices = v;
						// precalcola il vettore normale al poligono
						p1x = poly.vertices[0].x;	p1y = poly.vertices[0].y;	p1z = poly.vertices[0].z;
						pq.x = poly.vertices[1].x - p1x;		// calcola vettore pq
						pq.y = poly.vertices[1].y - p1y;
						pq.z = poly.vertices[1].z - p1z;
						pr.x = poly.vertices[v-1].x - p1x;		// calcola vettore pr
						pr.y = poly.vertices[v-1].y - p1y;
						pr.z = poly.vertices[v-1].z - p1z;
						Normalize(&pq);	// e' giusto?
						Normalize(&pr);	// e' giusto?
						// prodotto vettoriale -> vettore ortogonale al poligono
						CrossProduct(pq, pr, &n);
						Normalize(&n);
#if POLYGONS_CCW_IN_RML
						VectorMult(&n, -1.0);
#endif
						poly.dist = ((n.x * p1x) + (n.y * p1y) + (n.z * p1z));
						poly.normal.x = n.x; poly.normal.y = n.y; poly.normal.z = n.z;
						// immagazzina il nuovo poligono
						gStaticPolys[gStaticPolysCount] = poly;
						gStaticPolysCount++;
					}
				}
				err = GetNextTag(f, tagBuf);
			}
		} else if (strcmp(tagBuf, "map-sections") == 0) {		// map sections
			do { 
				err = GetNextTag(f, tagBuf);
				if (err != 0 || strcmp(tagBuf, "/map-sections") == 0)
					break;
				if (strncmp(tagBuf, "ms ", 3) == 0) {			// ms id_pavimento id_soffitto suono volume
					int	floorID, ceilID, soundID, soundVol;
					// attenzione a non confondere l'ID di un poligono
					// con il suo INDICE nella lista di poligoni.
					sscanf(tagBuf+3, "%d %d %d %d", &floorID, &ceilID, &soundID, &soundVol);
					NewMapSection(floorID, ceilID, soundID, soundVol, 0);
				}
			} while (1);
		} else if (strcmp(tagBuf, "objects") == 0) {	// inserimento di oggetti nella mappa
			do {
				err = GetNextTag(f, tagBuf);
				if (err != 0 || strcmp(tagBuf, "/objects") == 0)
					break;
				if (strncmp(tagBuf, "object", 6) == 0) {
					Point3D		pos = {0, 0, 0};
					Vector3D	speed = {0, 0, 0};
					float		rotx = 0, roty = 0, rotz = 0;
					int			model_id = -1;
					
					do {
						err = GetNextTag(f, tagBuf);
						if (strncmp(tagBuf, "pos ", 4) == 0)
							sscanf(tagBuf+4, "%f %f %f", &pos.x, &pos.y, &pos.z);
						else if (strncmp(tagBuf, "rot ", 4) == 0)
							sscanf(tagBuf+4, "%f %f %f", &rotx, &roty, &rotz);
						else if (strncmp(tagBuf, "model ", 6) == 0)
							model_id = GetModelIDFromName(tagBuf+6);
					} while (err == 0 && strcmp(tagBuf, "/object") != 0);
					if (model_id != NONE)
						NewObject(model_id, pos, speed, rotx, roty, rotz);
					else
						fprintf(stderr, "Ignoro un oggetto mal specificato\n");
				}
			} while (1);
		} else if (strncmp(tagBuf, "sky-color", 9) == 0) {	// colore di sfondo (meglio usare <sky><color r g b></sky>)
			sscanf(tagBuf+9, "%f %f %f", &clearColor[0], &clearColor[1], &clearColor[2]);
		} else if (strncmp(tagBuf, "sky", 3) == 0) {		// parametri cielo
			do {
				err = GetNextTag(f, tagBuf);
				if (err != 0 || strcmp(tagBuf, "/sky") == 0)
					break;
				if (strncmp(tagBuf, "color ", 6) == 0) {	// uguale a sky-color
					sscanf(tagBuf+6, "%f %f %f", &clearColor[0], &clearColor[1], &clearColor[2]);
				} else if (strncmp(tagBuf, "text", 4) == 0) {
					char	texfile[1024] = "";
					int		i = 0;

					while (1) {
						int	c = getc(f);

						if (c != '<')
							texfile[i++] = c;
						else
							break;
					}
					gSkyTexture = LoadTexture(texfile, ALPHAMODE_OPAQUE, GL_LINEAR);
				}
			} while (1);
		} else if (strncmp(tagBuf, "fog", 3) == 0) {	// parametri nebbia
			gFogParams.on = 1;
			do {
				err = GetNextTag(f, tagBuf);
				if (err != 0 || strcmp(tagBuf, "/fog") == 0)
					break;
				if (strncmp(tagBuf, "color ", 6) == 0)
					sscanf(tagBuf+6, "%f %f %f", &gFogParams.color[0], &gFogParams.color[1], &gFogParams.color[2]);
				else if (strncmp(tagBuf, "dist ", 5) == 0)
					sscanf(tagBuf+5, "%f",&gFogParams.dist);
			} while (1);
		} else if (strncmp(tagBuf, "sunlight", 8) == 0) {
			gGlobalLight.ambient[3] = 0.0;
			gGlobalLight.diffuse[3] = 0.0;
			do {
				err = GetNextTag(f, tagBuf);
				if (err != 0 || strcmp(tagBuf, "/sunlight") == 0)
					break;
				if (strncmp(tagBuf, "color ", 6) == 0)
					sscanf(tagBuf+6, "%f %f %f", &gGlobalLight.diffuse[0], &gGlobalLight.diffuse[1], &gGlobalLight.diffuse[2]);
				if (strncmp(tagBuf, "direction ", 10) == 0)
					sscanf(tagBuf+10, "%f %f %f", &gGlobalLight.dir.x, &gGlobalLight.dir.y, &gGlobalLight.dir.z);
				if (strncmp(tagBuf, "ambient ", 8) == 0)
					sscanf(tagBuf+8, "%f %f %f", &gGlobalLight.ambient[0], &gGlobalLight.ambient[1], &gGlobalLight.ambient[2]);
			} while (1);
		}
	} while (1);

	// Calcola le normali ai vertici di ogni poligono.
	// Per OGNI vertice del mondo, scopri tutti gli altri vertici
	// che hanno la stessa posizione, e fai la media tra le normali
	// dei loro poligoni di appartenenza: questa media e' la normale
	// del vertice, e ci serve per lo smooth shading.
	for (i = 0; i < gStaticPolysCount; i++) {
		int	vc;

		if (gStaticPolys[i].texture == NONE)
			continue;
		// per ogni vertice del poligono...
		for (vc = 0; vc < gStaticPolys[i].numVertices; vc++) {
#if CALC_VERTEX_NORMALS
			int			pc, sharedCount = 1;
			float		x = gStaticPolys[i].vertices[vc].x,
						y = gStaticPolys[i].vertices[vc].y,
						z = gStaticPolys[i].vertices[vc].z;
			Vector3D	n = gStaticPolys[i].normal;

			// per ogni poligono del mondo...
			for (pc = 0; pc < gStaticPolysCount; pc++) {
				int	vc2;
				
				// scarta alcuni tipi di poligono:
				// 1. questo stesso poligono :)
				// 2. poligoni senza texture
				// 3. poligoni di texture diversa da questo (dubbio)
				if (pc == i || gStaticPolys[pc].texture == NONE
					/*|| gStaticPolys[pc].texture != gStaticPolys[i].texture*/)
					continue;
				// per ogni vertice...
				for (vc2 = 0; vc2 < gStaticPolys[pc].numVertices; vc2++) {
					// se questo vertice e' "abbastanza vicino" al vertice
					// in esame, aggiungi la normale alla media
					if (fabsf(gStaticPolys[pc].vertices[vc2].x - x) < SMALL_DIFF &&
						fabsf(gStaticPolys[pc].vertices[vc2].y - y) < SMALL_DIFF &&
						fabsf(gStaticPolys[pc].vertices[vc2].z - z) < SMALL_DIFF) {
						n.x += gStaticPolys[pc].normal.x;
						n.y += gStaticPolys[pc].normal.y;
						n.z += gStaticPolys[pc].normal.z;
						sharedCount++;
					}
				}
			}
			// normalizza il vettore e mettilo via
			Normalize(&n);
			gStaticPolys[i].vertices[vc].nx = n.x;
			gStaticPolys[i].vertices[vc].ny = n.y;
			gStaticPolys[i].vertices[vc].nz = n.z;
#else
			gStaticPolys[i].vertices[vc].nx = gStaticPolys[i].normal.x;
			gStaticPolys[i].vertices[vc].ny = gStaticPolys[i].normal.y;
			gStaticPolys[i].vertices[vc].nz = gStaticPolys[i].normal.z;
#endif
		}
	}

	// costruisci le ombre portate della luce del sole, rispetto al mondo.
	// Queste vengono realizzate sovrapponendo dei poligoni senza texture
	// alle parti in ombra, e utilizzando quindi lo stencil buffer per
	// oscurare solo queste zone coperte. Non va benissimo.
	fprintf(stderr, "Costruisco le lightmaps... ");
	for (i = 0; i < gStaticPolysCount; i++) {
		Vector3D		n2, n;	// normale al piano su cui giace il poligono corrente
		Polygon3D		*currPoly = &gStaticPolys[i];
		float			dist, dot;
		unsigned int	j, shadowCount = 0;
		Vector3D		faceNorms[kMaxVertices];
		float			faceDists[kMaxVertices];
		ShadowPoly		sbuf[128];	// luogo temporaneo per tenere le ombre

		if (currPoly->texture == NONE)
			continue;
		n = currPoly->normal;
		dist = currPoly->dist;
		// controlla se il poligono e' orientato nella stessa dir della luce.
		// Se e' cosi', e' completamente in ombra e non serve calcolare ombre portate.
		dot = DotProduct(n, gGlobalLight.dir);
		if (dot >= 0)
			goto bl;

		// calcola i piani perpendicolari al poligono corrente e passanti
		// per i suoi lati - in pratica, il "prisma" che ha come base questo poligono.
		// In seguito, elimineremo tutte le ombre che cadono fuori da questo prisma.
		n2 = n;
#if POLYGONS_CCW_IN_RML
		VectorMult(&n2, -1);
#endif
		for (j = 0; j < currPoly->numVertices; j++) {
			Vector3D	ab;
			
			if (j == currPoly->numVertices - 1) {
				ab.x = currPoly->vertices[0].x - currPoly->vertices[j].x;
				ab.y = currPoly->vertices[0].y - currPoly->vertices[j].y;
				ab.z = currPoly->vertices[0].z - currPoly->vertices[j].z;
			} else {
				ab.x = currPoly->vertices[j+1].x - currPoly->vertices[j].x;
				ab.y = currPoly->vertices[j+1].y - currPoly->vertices[j].y;
				ab.z = currPoly->vertices[j+1].z - currPoly->vertices[j].z;
			}
			CrossProduct(n2, ab, &faceNorms[j]);
			Normalize(&faceNorms[j]);
			faceDists[j] = DotProduct(currPoly->vertices[j], faceNorms[j]);
		}

		// costruisci le ombre generate dagli altri poligoni
		for (j = 0; j < gStaticPolysCount; j++) {
			Polygon3D	*poly = &gStaticPolys[j];
			ShadowPoly	shadow;
			int			vert, noshadow = 0;
			
			if (j == i || poly->texture == NONE)
				continue;
			shadow.numVertices = poly->numVertices;
			// proietta ogni vertice sul piano lungo la direzione della luce (uffa)
			for (vert = 0; vert < poly->numVertices; vert++) {
				Point3D	v;
				float	alfa, div;

				v.x = poly->vertices[vert].x;
				v.y = poly->vertices[vert].y;
				v.z = poly->vertices[vert].z;
				div = DotProduct(n, gGlobalLight.dir);
				if (div == 0.0)
					alfa = 1e4;	// luce parallela al piano: proiezione infinita (con l'abbso)
				else
					alfa = (dist - DotProduct(n, v)) / div;
				// controlla che nessun vertice con proiezione negativa
				// cada dentro al poligono... sarebbe un errore
				if (alfa < 0) {
					int	foo;
					
					for (foo = 0; foo < currPoly->numVertices; foo++) {
						if (DotProduct(v, faceNorms[foo]) > faceDists[foo])
							noshadow = 1;
					}
				}
				shadow.vertices[vert].x = v.x + alfa * gGlobalLight.dir.x;
				shadow.vertices[vert].y = v.y + alfa * gGlobalLight.dir.y;
				shadow.vertices[vert].z = v.z + alfa * gGlobalLight.dir.z;
			}
			if (!noshadow && (shadowCount < 128)) {
				// prima di mettere via l'ombra, ritagliala
				// contro il poligono su cui e' proiettata
				// (non devono cadere fuori pezzi di ombra :)
				unsigned int	facei;

				for (facei = 0; facei < currPoly->numVertices; facei++) {
					ShadowPoly	newShadow;

					ClipShadowToPlane(&shadow, &faceNorms[facei], faceDists[facei], &newShadow);
					shadow = newShadow;
				}
				if (shadow.numVertices > 2)
					sbuf[shadowCount++] = shadow;
			}
		}
bl:
		if (shadowCount != 0)
			hireslmcount++;
		BuildLightmap(currPoly, sbuf, shadowCount);
	}
	fprintf(stderr, "fatto: %d lightmap hires (%.0f%%)\n", hireslmcount, 100.0*hireslmcount/gStaticPolysCount);
	
	gCamera.ySpeed = 0.0;	
	return 0;
}

// ritaglia un'ombra in modo che copra solo il suo poligono
static void ClipShadowToPlane(ShadowPoly *in, Vector3D *normal, float dist, ShadowPoly *out)
{
	unsigned int	i, outVertCount = 0;
	Vector3D		n = *normal;

	for (i = 0; i < in->numVertices; i++) {
		unsigned int	bi = (i == in->numVertices - 1) ? 0 : i + 1;
		Point3D			a = in->vertices[i], b = in->vertices[bi];
		float			dotA, dotB;

		dotA = DotProduct(a, n);
		dotB = DotProduct(b, n);
		if (dotA > dist) {	// a cade dentro
			if (dotB > dist) {	// anche b
				// salva a e b
				if (i == 0)		out->vertices[outVertCount++] = a;
				if (bi != 0)	out->vertices[outVertCount++] = b;
			} else { // b cade fuori
				// salva a e intersezione
				float	div = (dist - dotA) / (dotB - dotA);
				Point3D	inters;
				
				if (i == 0)	out->vertices[outVertCount++] = a;
				inters.x = a.x + div * (b.x - a.x);
				inters.y = a.y + div * (b.y - a.y);
				inters.z = a.z + div * (b.z - a.z);
				out->vertices[outVertCount++] = inters;
			}
		} else {	// a cade fuori
			if (dotB > dist) {	// b cade dentro
				// salva intersezione e b
				float   div = (dist - dotA) / (dotB - dotA);
				Point3D inters;

				inters.x = a.x + div * (b.x - a.x);
				inters.y = a.y + div * (b.y - a.y);
				inters.z = a.z + div * (b.z - a.z);
				out->vertices[outVertCount++] = inters;
				if (bi != 0)	out->vertices[outVertCount++] = b;
			}
		}
	}
	out->numVertices = outVertCount;
}

// aggiungi una map section alla lista, partendo da tutti i dati necessari (poligono pavimento, soffitto eccetera eccetera)
void NewMapSection(int floorID, int ceilID, int sound, long sVolume, float liquidY)
{
	MapSection	ms;
	int			i, floor_poly = NONE, ceil_poly = NONE;
	
	// scova gli indici dei poligoni soffitto e pavimento
	for (i = 0; i < gStaticPolysCount; i++) {
		if (gStaticPolys[i].id == floorID)
			floor_poly = i;
		else if (gStaticPolys[i].id == ceilID)
			ceil_poly = i;
		if (floor_poly != NONE && ceil_poly != NONE)
			break;
	}
	ms.numVertices = gStaticPolys[floor_poly].numVertices;
	// copia x e z dei vertici del pavimento nella map section
	for (i = 0; i < ms.numVertices; i++) {
		ms.vertices[i].x = gStaticPolys[floor_poly].vertices[i].x;
		ms.vertices[i].y = gStaticPolys[floor_poly].vertices[i].z;
	}
	// adesso calcola i vettori normali ai lati della map section, ci serviranno per il collision detect
	for (i = 0; i < ms.numVertices; i++) {
		Vector2D	n;
		float		pq_x, pq_z;
		float		pr_x, pr_y, pr_z;
		float		p1x, p1z;
		int			i2 = 0;
		
		p1x = ms.vertices[i].x;				// prendi un vertice
		p1z = ms.vertices[i].y;
		if (i != (ms.numVertices - 1))
			i2 = i + 1;
		pq_x = p1x - ms.vertices[i2].x;		// calcola il vettore da questo vertice al prossimo
		pq_z = p1z - ms.vertices[i2].y;
		pr_x = 0;							// e' sufficiente un vettore verticale di lunghezza 1,
		pr_y = 1;							// perche' sappiamo che la normale sara' orizzontale
		pr_z = 0;							// (diamo per scontato che la parete sia verticale)
		n.x = -pq_z;						// calcola il prodotto scalare tra i due vettori -> ecco la nostra normale
		n.y = pq_x;
		Normalize2D(&n);					// rendi il vettore ottenuto di lunghezza 1
		ms.wallsNormals[i] = n;
		ms.wallsDists[i] = (n.x * p1x) + (n.y * p1z);
	}
	// salva altezze di pavimento e soffitto
	ms.floorY = gStaticPolys[floor_poly].vertices[0].y;
	if (ceil_poly != NONE)
		ms.ceilingY = gStaticPolys[ceil_poly].vertices[0].y;
	else
		ms.ceilingY = ms.floorY;	// significa che il soffitto non c'e' (non so quanto questo possa essere utile, forse e' meglio toglierlo)
	ms.liquidHeight = liquidY;
	ms.floorPolyID = floor_poly;
	ms.ceilingPolyID = ceil_poly;
	ms.ambientSound = sound;
	ms.ambientVolume = sVolume;
	
	gMap[gMapSectionsNum] = ms;		// memorizza nella lista questa stanza
	gMapSectionsNum++;
}

/**** COLLISION DETECTION ****/

// verifica se un punto 3D e' dentro una map section
int PointInMapSection(MapSection *ms, float px, float py, float pz)
{
	int	i;
	
	if (py <= ms->floorY)				// se il punto e' sotto al pavimento, esci
		return 0;
	if (ms->ceilingY > ms->floorY) {	// se questa section ha il soffitto
		if (py >= ms->ceilingY)			// ...e il punto e' sopra di esso, esci
			return 0;
	}
	// adesso bisogna controllare se px e py sono all'interno delle pareti
	// della section. Basta un semplice prodotto scalare...
	for (i = 0; i < ms->numVertices; i++) {
		float	prod;
		
		prod = (px * ms->wallsNormals[i].x) + (pz * ms->wallsNormals[i].y);
		if (prod > ms->wallsDists[i])
			return 0;
	}
	return 1;
}

// ritorna l'indice del lato della map section intersecato dal vettore.
// FIXME: poco affidabile se interseca la section vicino ad un vertice.
int IntersectedSectionSide(int section, float x1, float y1, float z1, float x2, float y2, float z2)
{
	int			i, result = NONE;
	MapSection	ms = gMap[section];
	
	if (y1 > ms.floorY && y2 < ms.floorY)
		return FLOOR;
	if (y1 < ms.ceilingY && y2 > ms.ceilingY)
		return CEILING;
	for (i = ms.numVertices-1; i >= 0; i--) {
		float	prod1, prod2, dist;

		dist = ms.wallsDists[i];
		prod1 = (x1 * ms.wallsNormals[i].x) + (z1 * ms.wallsNormals[i].y);
		prod2 = (x2 * ms.wallsNormals[i].x) + (z2 * ms.wallsNormals[i].y);
		// il primo e' dentro e il secondo fuori
		if (prod1 <= dist && prod2 >= dist)
			result = i;
	}
	return result;
}

// controlla se una sfera e' all'interno di una section; se non lo e',
// ritorna in hitSide il lato che interseca.
// FIXME: e se ne interseca piu' di uno?? eeh??
int SphereInMapSection(int ms, float x, float y, float z, float radius, int *hitSide)
{
	int			i;
	MapSection	*msp = &gMap[ms];

	if (y - radius < msp->floorY) {
		*hitSide = FLOOR;
		return 0;
	}
	if (y + radius > msp->ceilingY) {
		*hitSide = CEILING;
		return 0;
	}
	for (i = msp->numVertices-1; i >= 0; i--) {
		float	prod, dist = msp->wallsDists[i];

		prod = (x * msp->wallsNormals[i].x) + (z * msp->wallsNormals[i].y);
		if (prod + radius > dist) {
			*hitSide = i;
			return 0;
		}
	}
	*hitSide = NONE;
	return 1;
}

// modifica io_delta in modo da restare dentro la map section specificata
// (scivolando sui lati contro cui sbattiamo).
void SlideInsideMapSection(int ms, float x1, float y1, float z1, Vector3D *io_delta)
{
	int 	side;
	float	x2 = x1 + io_delta->x,
			y2 = y1 + io_delta->y,
			z2 = z1 + io_delta->z;
	
	side = IntersectedSectionSide(ms, x1, y1, z1, x2, y2, z2);
	if (side != NONE) {
		Vector3D    normal = {0, 0, 0}, up, delta;
		
		if (side == FLOOR) {
			normal.y = 1;
		} else if (side == CEILING) {
			normal.y = -1;
		} else {
			normal.x = gMap[ms].wallsNormals[side].x;
			normal.z = gMap[ms].wallsNormals[side].y;
		}
		delta = *io_delta;
		CrossProduct(normal, delta, &up);
		CrossProduct(normal, up, &delta);
		io_delta->x = -delta.x;
		io_delta->y = -delta.y;
		io_delta->z = -delta.z;
		// controlla se con il nuovo spostamento siamo andati fuori dalla section
		// FIXME: in realta' qui bisognerebbe rifare tutto il rilevamento delle
		//        collisioni, ma diventerebbe una cosa infinita...
		if (!PointInMapSection(&gMap[ms],
					x1+io_delta->x,
					y1+io_delta->y,
					z1+io_delta->z)) {
			// si, annulla lo spostamento
			io_delta->x = 0;
			io_delta->y = 0;
			io_delta->z = 0;
		}
	} else {
		io_delta->x = io_delta->y = io_delta->z = 0;
	}
}

// come prima, ma forniamo noi il lato colpito
// (serve per scivolare rispettando il raggio della camera)
void SlideInsideMapSection2(int ms, float x1, float y1, float z1, int side, Vector3D *io_delta)
{
	if (side != NONE) {
		Vector3D    normal = {0, 0, 0}, up, delta;

		if (side == FLOOR) {
			normal.y = 1;
		} else if (side == CEILING) {
			normal.y = -1;
		} else {
			normal.x = gMap[ms].wallsNormals[side].x;
			normal.z = gMap[ms].wallsNormals[side].y;
		}
		delta = *io_delta;
		CrossProduct(normal, delta, &up);
		CrossProduct(normal, up, &delta);
		io_delta->x = -delta.x;
		io_delta->y = -delta.y;
		io_delta->z = -delta.z;

		if (!PointInMapSection(&gMap[ms],
					x1+io_delta->x,
					y1+io_delta->y,
					z1+io_delta->z)) {
			io_delta->x = 0;
			io_delta->y = 0;
			io_delta->z = 0;
		}
	} else {
		io_delta->x = io_delta->y = io_delta->z = 0;
	}
}

