/*******************************************************
  Rockland >> Gestione degli oggetti presenti nel mondo
 *******************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include "global.h"
#include "worlddata.h"
#include "objects.h"
#include "textures.h"
#include "map.h"
#include "3d-calc.h"

static void FireballExplosion(struct Object *o);

Object				*gLastObject = NULL, *gFirstObject = NULL;
int					gObjectCount;
int					gDrawnObjects;
extern Polygon3D	gStaticPolys[];
extern int			gMapSectionsNum;
extern MapSection	gMap[];
extern Camera		gCamera;
extern FogParams	gFogParams;
extern LightSource	gGlobalLight;

GLUquadricObj		*quadr, *ballQuadr, *notexquadr;

GLuint			tex_antinazi = 0;
GLuint			tex_football = 0;
GLuint			tex_flames = 0;
GLuint			tex_gate_icon = 0;
GLuint			tex_kerosene = 0;
GLuint			tex_geek = 0, envmap2 = 0, envmap3 = 0;
GLuint			tex_doge = 0, tex_piantine[3] = {0, 0, 0};
GLint			gLampCounter = 0, gGrassCounter = 0;
extern GLuint	textures[];

void InitObjects(void)
{
	fprintf(stderr, "Inizializzo il gestore degli oggetti...");
	
	gObjectCount = 0;
	gLastObject = NULL;
	gFirstObject = NULL;
	quadr = gluNewQuadric();
	gluQuadricNormals(quadr, GLU_SMOOTH);
	gluQuadricTexture(quadr, GL_TRUE);
	ballQuadr = gluNewQuadric();
	gluQuadricNormals(ballQuadr, GLU_SMOOTH);
	gluQuadricTexture(ballQuadr, GL_TRUE);
	notexquadr = gluNewQuadric();
	
	tex_flames = LoadTexture("textures/flames.bmp", ALPHAMODE_COLOR_AVERAGE, GL_LINEAR);
	tex_geek = LoadTexture("textures/envmap.jpg", ALPHAMODE_OPAQUE, GL_LINEAR);
	envmap2 = LoadTexture("textures/envmap-metal.jpg", ALPHAMODE_OPAQUE, GL_LINEAR);
	envmap3 = LoadTexture("textures/envmap3.jpg", ALPHAMODE_OPAQUE, GL_LINEAR);
	
	fprintf(stderr, " fatto\n");
}

void NewObject(int model, Point3D start_pos, Vector3D start_speed, float ax, float ay, float az)
{
	Object	*new = NULL;
	int		i;

	new = (Object *)malloc(sizeof(Object));
	if (new == NULL)
		return;
	// inizializza i dati del nuovo oggetto
	new->model = model;
	new->flags = 0;
	new->other = 0;
	new->position = start_pos;
	new->speed = start_speed;
	new->pitch = ax;
	new->yaw = ay;
	new->roll = az;
	new->decay = 1.0;
	new->destroyProc = NULL;
	new->collDetDone = 0;
	switch (model) {
		case MODEL_FIREBALL:
			new->radius = 0.05;
			new->destroyProc = FireballExplosion;
			break;
		case MODEL_SMOKEPUFF:
			new->radius = 0.01;
			new->flags |= FLAGS_2D;
			break;
		case MODEL_BOUNCING:
			new->radius = 0.1;
			new->flags |= SOLID;
			if (tex_football == 0)
				tex_football = LoadTexture("textures/palla.bmp", ALPHAMODE_OPAQUE, GL_LINEAR);
			break;
		case MODEL_PARTICLETORCH:
			new->radius = 0.1;
			new->flags |= IS_SWITCH;
			new->flags |= SOLID | FIXED;
			new->other |= SWITCH_ON_MASK;
			if (tex_gate_icon == 0)
				tex_gate_icon = LoadTexture("textures/segnale-cancello.bmp", ALPHAMODE_OPAQUE, GL_LINEAR);
			break;
		case MODEL_TANK:
			new->radius = 0.25;
			new->flags |= SOLID;
			if (tex_kerosene == 0)
				tex_kerosene = LoadTexture("textures/kerosene.jpg", ALPHAMODE_OPAQUE, GL_LINEAR);
			break;
		case MODEL_HANDRAIL:
			new->radius = 0.01;
			new->flags |= SOLID | FIXED;
			break;
		case MODEL_BRICK:
			new->radius = 0.09;
			new->flags |= SOLID | FIXED;
			break;
		case MODEL_DOGE:
			new->radius = 0.4;
			new->flags |= SOLID | FLAGS_2D;
			if (tex_doge == 0)
				tex_doge = LoadTexture("textures/doge.bmp", ALPHAMODE_GREEN, GL_LINEAR);
			break;
		case MODEL_ANTINAZI:
			if (tex_antinazi == 0)
				tex_antinazi = LoadTexture("textures/antinazi.bmp", ALPHAMODE_GREEN, GL_LINEAR);
			break;
		case MODEL_LAMP:
			new->radius = 0.3;
			new->flags |= SOLID | FIXED;
			new->other = (unsigned long)(GL_LIGHT2 + gLampCounter);
			gLampCounter++;
			break;
		case MODEL_GRASS:
			new->radius = 0.2;
			new->flags |= FIXED | FLAGS_2D;
			new->other = gGrassCounter;
			gGrassCounter++;
			if (gGrassCounter > 2)
				gGrassCounter = 0;
			if (tex_piantine[new->other] == 0) {
				char	fname[1024] = "";

				sprintf(fname, "textures/piantina%d.bmp", new->other + 1);
				tex_piantine[new->other] = LoadTexture(fname, ALPHAMODE_BLUE, GL_LINEAR);
			}
			break;
		default:
			new->radius = 0.0;
			break;
	}
	new->mapSection = -1;
	for (i = 0; i < gMapSectionsNum; i++) {
		if (PointInMapSection(&gMap[i], start_pos.x, start_pos.y, start_pos.z)) {
			new->mapSection = i;
			break;
		}
	}
	if (new->mapSection == -1) {	// niente oggetti fuori dal mondo
		free(new);
		return;
	}
	new->next = NULL;
	new->prev = NULL;
	if (gObjectCount == 0) {
		gFirstObject = new;
		gLastObject = new;
	} else {
		new->prev = gLastObject;
		gLastObject->next = new;
		gLastObject = new;
	}
	gObjectCount++;
}

// scollega l'oggetto dalla lista linkata e liberane la memoria
void DestroyObject(Object *o)
{
	Object	*next, *prev;
	
	// questo serve per fare qualcosa quando questo oggetto
	// viene distrutto (p.es. quando un oggetto "missile" viene
	// distrutto, destroyProc crea un'oggetto "esplosione").
	if (o->destroyProc != NULL)
		(o->destroyProc)(o);
	
	next = o->next;
	prev = o->prev;
	// scollega l'oggetto dalla lista linkata
	if (prev != NULL) {
		prev->next = next;
	} else {
		// era il primo oggetto della catena
		gFirstObject = next;
		if (next != NULL)
			next->prev = NULL;
	}
	if (next != NULL) {
		next->prev = prev;
	} else {
		// era l'ultimo oggetto della catena
		gLastObject = prev;
		if (prev != NULL)
			prev->next = NULL;
	}
	// dealloca l'oggetto
	free(o);
	gObjectCount--;
}

// Genera un'esplosione quando viene deallocato un oggetto fireball
static void FireballExplosion(struct Object *o)
{
	Vector3D	speed = {0, 0, 0};
	Object		*co = gFirstObject;

	// applica una forza a tutti gli oggetti palla nelle vicinanze,
	// in modo da scagliarli lontano dall'esplosione
	while (co != NULL) {
		if (co->model == MODEL_BOUNCING || co->model == MODEL_SMOKEPUFF || co->model == MODEL_DOGE) {
			Vector3D	dist;
			float		distdist;

			dist.x = co->position.x - o->position.x;
			dist.y = co->position.y - o->position.y;
			dist.z = co->position.z - o->position.z;
			distdist = (dist.x * dist.x) + (dist.y * dist.y) + (dist.z * dist.z);
			Normalize(&dist);
			// evita distdist = 0: verrebbe applicata una forza infinita :)
			if (distdist < 0.0005)
				distdist = 0.0005;
			if (distdist < 5.0) {
				co->speed.x += dist.x * 0.15 / sqrt(distdist);
				co->speed.y += dist.y * 0.15 / sqrt(distdist);
				co->speed.z += dist.z * 0.15 / sqrt(distdist);
			}
		}
		co = co->next;
	}
}

void UpdateObjects(double dt)
{
	Object	*o = gFirstObject;

	dt *= 0.05;
	while (o != NULL) {
		Vector3D	speed = o->speed;
		Object		*temp = o->next;
		float		x = o->position.x, y = o->position.y, z = o->position.z;
		float		radius = o->radius;
		int			escapingWorld = 0, escapeMSSide = -1;

		x += speed.x * dt;
		y += speed.y * dt;
		z += speed.z * dt;
		if (o->model == MODEL_SMOKEPUFF) {
			x += cos(z * 3) * 0.008 * (y-1);
			z += sin(y * 3) * 0.008 * (y-1);
		}

		// rileva le collisioni con il mondo
		if (!SphereInMapSection(o->mapSection, x, y, z, radius, &escapeMSSide)) {
			// l'oggetto non e' piu' nella sua vecchia section, nel senso che
			// la sfera circoscritta ad esso non e' completamente dentro la section.
			escapingWorld = 1;
			// se la direzione e' "simile alla normale della parete",
			// possiamo dire che l'oggetto sta tornando dentro la section,
			// quindi non fermarlo.
			if (escapeMSSide > -1) {
				Vector2D	normal = gMap[o->mapSection].wallsNormals[escapeMSSide];
				Vector2D	direction;
				float		dot;

				direction.x = speed.x;
				direction.y = speed.z;
				dot = (normal.x * direction.x) + (normal.y * direction.y);
				if (dot < 0)
					escapingWorld = 0;
			}
			if (!PointInMapSection(&gMap[o->mapSection], x, y, z)) {
				// il centro dell'oggetto non e' piu' nella map section.
				// Ok, l'oggetto e' totalmente fuori.
				int	i, newMapSection = -1;
			
				// cerca in quale section si trova ora il centro
				for (i = 0; i < gMapSectionsNum; i++) {
					if (PointInMapSection(&gMap[i], x, y, z) && i != o->mapSection) {
						newMapSection = i;
						break;
					}
				}
				if (newMapSection != NONE) {
					// trovata, memorizzala come map section corrente
					o->mapSection = newMapSection;
					escapingWorld = 0;	// non sta uscendo dal mondo, ha cambiato section
				} else {
					// l'oggetto e' uscito completamente dal mondo
					escapingWorld = 2;
					// calcola quale lato della map section ha colpito
					// (ci serve per variare la sua direzione ecc.)
					escapeMSSide = IntersectedSectionSide(o->mapSection,
									o->position.x, o->position.y, o->position.z,
									x, y, z);
				}
			} else if (escapeMSSide > -1) {
				// il centro e' ancora nella vecchia section! Dobbiamo
				// controllare se stiamo attraversando il confine tra due section
				Vector2D	v;
				float		test_x, test_y, test_z;
				int			i;

				// calcola un punto della sfera che sicuramente e' fuori
				// dalla section corrente...
				v = gMap[o->mapSection].wallsNormals[escapeMSSide];
				v.x *= -radius;
				v.y *= -radius;
				test_x = -v.x + x;
				test_y = y;
				test_z = -v.y + z;
				// ...quindi controlla se questo punto appartiene
				// ad una section diversa da quella corrente
				for (i = 0; i < gMapSectionsNum; i++) {
					if (i != o->mapSection && PointInMapSection(&gMap[i], test_x, test_y, test_z)) {
						// si, appartiene alla section i.
						// L'oggetto non sembra colpire ostacoli, lascialo proseguire
						escapingWorld = 0;
						break;
					}
				}
			}
		}
		
		// collisioni con altri oggetti
		if (o->flags & SOLID) {
			Object	*mate = gFirstObject;
			
			while (mate != NULL) {
				float	dx, dy, dz, dist, radSum;
				
				if (mate == o || !(mate->flags & SOLID) || mate->collDetDone) {
					mate = mate->next;
					continue;
				}
				
				dx = mate->position.x - x;
				dy = mate->position.y - y;
				dz = mate->position.z - z;
				dist = dx * dx + dy * dy + dz * dz;
				radSum = mate->radius + radius;
				if (dist < (radSum * radSum)) {
					Vector3D	normal = {dx, dy, dz},
							totalSpeed, s1f, s2f;
					float		dot;
					
					Normalize(&normal);
					
					/* calcola le velocita' finali
					   FIXME: questa equazione E' SBAGLIATA, ma dato
					   che piu' o meno funziona, per ora uso questa :)
					   Il problema e' che, secondo la fisica classica,
					   NON e' possibile risolvere analiticamente il sistema
					   dell'urto elastico di due corpi nello spazio a tre
					   dimensioni. Quindi... come diavolo facciamo?? */
					totalSpeed.x = (mate->speed.x + o->speed.x);
					totalSpeed.y = (mate->speed.y + o->speed.y);
					totalSpeed.z = (mate->speed.z + o->speed.z);
					dot = 2.0 * ((-totalSpeed.x * normal.x)
							+ (-totalSpeed.y * normal.y)
							+ (-totalSpeed.z * normal.z));
					s1f.x = (dot * normal.x + totalSpeed.x) * 0.4;
					s1f.y = (dot * normal.y + totalSpeed.y) * 0.4;
					s1f.z = (dot * normal.z + totalSpeed.z) * 0.4;
					s2f.x = (-dot * normal.x + totalSpeed.x) * 0.4;
					s2f.y = (-dot * normal.y + totalSpeed.y) * 0.4;
					s2f.z = (-dot * normal.z + totalSpeed.z) * 0.4;
					
					if (!(o->flags & FIXED))
						o->speed = s1f;
					if (!(mate->flags & FIXED))
						mate->speed = s2f;
					o->collDetDone = 1;
					mate->collDetDone = 1;
					break;
				}
				
				mate = mate->next;
			}
		}
	
		// azioni specifiche per i vari tipi di oggetto
		switch (o->model) {
			case MODEL_FIREBALL:
				o->speed.y -= 0.005 * dt;
				if (escapingWorld) {
					Vector3D		speed = {0, 0, 0};
					Vector3D		normal = {0, 0, 0};
					float			x_angle, y_angle, z_angle;
					
					// calcola la normale alla parete colpita
					if (escapeMSSide == FLOOR)
						normal.y = 1;
					else if (escapeMSSide == CEILING)
						normal.y = -1;
					else {
						normal.x = gMap[o->mapSection].wallsNormals[escapeMSSide].x;
						normal.z = gMap[o->mapSection].wallsNormals[escapeMSSide].y;
					}
					// ora, in base alla normale, calcola gli angoli
					// x e y per orientare correttamente l'esplosione
					// FIXME
					y_angle = atan2(normal.x, normal.z) * 180.0 / M_PI;
					if (normal.z > 0)
						y_angle = -y_angle;
					x_angle = atan2(normal.y, normal.z) * 180.0 / M_PI;
					z_angle = atan2(normal.y, normal.z) * 180.0 / M_PI;
					NewObject(MODEL_EXPLOSION, o->position, speed, x_angle, y_angle, z_angle);	
					DestroyObject(o);
				}
				break;
			case MODEL_EXPLOSION:
				o->decay -= 0.03 * dt;
				o->radius = 1.06 - o->decay;
				if (o->decay <= 0.0) {
					glDisable(GL_LIGHT1);
					DestroyObject(o);
				}
				break;
			case MODEL_SMOKEPUFF:
				o->decay -= 0.017 * dt;
				o->radius = 0.4 - (o->decay * 0.4);
				if (o->decay < 0.0 || (escapingWorld && escapeMSSide != CEILING))
					DestroyObject(o);
				break;
			case MODEL_PARTICLETORCH:
				{
					unsigned long	animCounter;
					Point3D		pos = o->position;
					Vector3D	speed = {((float)random() / RAND_MAX - 0.5) * 0.006,
										((float)random() / RAND_MAX) * 0.013 + 0.01,
										((float)random() / RAND_MAX - 0.5) * 0.006};

					if (o->other & SWITCH_ON_MASK) {
						pos.y += 1.0;
						NewObject(MODEL_SMOKEPUFF, pos, speed, 0, 0, 0);
					}
					animCounter = (o->other & SWITCH_ANIM_MASK) >> 24;
					if (animCounter <= 70) {
						animCounter++;
						o->other &= SWITCH_ON_MASK;
						o->other |= animCounter << 24;
					}
				}
				break;
			case MODEL_BOUNCING:
				o->pitch -= o->speed.z * 350.0 * dt;
				if (fabsf(o->pitch) > 359.0)
					o->pitch = 0.0;
				o->roll += o->speed.x * 350.0 * dt;
				if (fabsf(o->roll) > 359.0)
					o->roll = 0.0;
				// rimbalza contro le pareti
				if (escapingWorld) {
					if (escapeMSSide == FLOOR) {
						y = gMap[o->mapSection].floorY + radius;
						o->speed.y = -o->speed.y * 0.75;
						o->speed.x *= 0.93;
						o->speed.z *= 0.93;
					} else if (escapeMSSide == CEILING) {
						y = gMap[o->mapSection].ceilingY - radius;
						o->speed.y = -o->speed.y * 0.75;
						o->speed.x *= 0.93;
						o->speed.z *= 0.93;
					} else if (escapeMSSide != NONE) {	// parete
						Vector2D    normal; // _|_ al muro colpito
						float       dot;
					
						normal = gMap[o->mapSection].wallsNormals[escapeMSSide];
						// new_speed = 2(-speed . normal) x normal + speed
						dot = 2.0 * ((-speed.x * normal.x) + (-speed.z * normal.y));
						o->speed.x = (dot * normal.x + speed.x) * 0.9;
						o->speed.z = (dot * normal.y + speed.z) * 0.9;
					} else {
						// per sicurezza, se un oggetto si "perde"
						// lo DISTRUGGIAMO IMMEDIATAMENTE
						DestroyObject(o);
					}
				} else {
					o->speed.y -= 0.002 * dt;
				}
				break;
			case MODEL_FOOBAR:
				o->pitch += 0.4 * dt;
				if (o->pitch > 359.0)	o->pitch = 0.0;
				o->roll += 0.48 * dt;
				if (o->roll > 359.0)	o->roll = 0.0;
				o->yaw += 0.36 * dt;
				if (o->yaw > 359.0)	o->yaw = 0.0;
				break;
			case MODEL_DOGE:
				if (escapingWorld) {
					if (escapeMSSide == FLOOR || escapeMSSide == CEILING) {
						o->speed.y = -o->speed.y * 0.94;
						o->speed.x *= 0.95;
						o->speed.z *= 0.95;
						o->position.y = (escapeMSSide == FLOOR) ? gMap[o->mapSection].floorY : gMap[o->mapSection].ceilingY;
						o->position.y += o->radius;
					} else if (escapeMSSide != NONE) {  // parete
						Vector2D    normal; // _|_ al muro colpito
						float       dot;

						normal = gMap[o->mapSection].wallsNormals[escapeMSSide];
						// new_speed = 2(-speed . normal) x normal + speed
						dot = 2.0 * ((-speed.x * normal.x) + (-speed.z * normal.y));
						o->speed.x = 0.99 * (dot * normal.x + speed.x);
						o->speed.z = 0.99 * (dot * normal.y + speed.z);
					}
				} else {
					o->speed.y -= 0.002 * dt;
				}
				if ((SDL_GetTicks() % 10) == 0) {
					o->speed.x += (((float)random() / (float)RAND_MAX) - 0.5) * 0.02;
					o->speed.y += (((float)random() / (float)RAND_MAX) - 0.5) * 0.006;
					o->speed.z += (((float)random() / (float)RAND_MAX) -0.5) * 0.02;
				}
				break;
			default:
				break;
		}
		if (!escapingWorld) {
			o->position.x = x;
			o->position.y = y;
			o->position.z = z;
		}
		
nextObj:
		o = temp;
	}
}

// invia un'azione all'oggetto "agibile" piu'
// vicino al giocatore (e non troppo lontano ovviamente)
void BroadcastObjectAction(void)
{
	Object	*o = gFirstObject;
	Point3D	pp = {gCamera.x, gCamera.y, gCamera.z};

	while (o != NULL) {
		float	dx, dy, dz, d;
	
		if (!(o->flags & IS_SWITCH))
			goto skipThis;

		// ricava la distanza tra noi e l'oggetto
		dx = pp.x - o->position.x;
		dy = pp.y - o->position.y;
		dz = pp.z - o->position.z;
		d = dx * dx + dy * dy + dz * dz;
		if (d > (OBJECT_ACTION_MAX_DIST * OBJECT_ACTION_MAX_DIST))
			goto skipThis;	// e' troppo lontano
		
		// possiamo azionarlo...
		if (o->other & SWITCH_ON_MASK) {
			o->other &= ~SWITCH_ANIM_MASK;	// avvia l'animazione
			o->other &= ~SWITCH_ON_MASK;
		} else {
			o->other &= ~SWITCH_ANIM_MASK;	// avvia l'animazione
			o->other |= SWITCH_ON_MASK;
		}
		return;
skipThis:
		o = o->next;
	}
}

void RenderObjects(void)
{
	Object  *o = gFirstObject;

	gDrawnObjects = 0;
	while (o != NULL) {
		o->collDetDone = 0;
		// calcola se dobbiamo disegnare l'oggetto o possiamo tralasciarlo
		if (o->model != MODEL_LAMP) {
			float	dx, dz, angle, delta, camera;

			// calcola il coeff. angolare del vettore camera-oggetto
			dx = o->position.x - gCamera.x;
			dz = o->position.z - gCamera.z;
			// ora esamina i vari casi del segno di questo coefficiente,
			// e ricava gli angoli corrispondenti (e' geometria, ragazzo)
			if (dx > 0) {
				if (dz >= 0)
					angle = atan(dz / dx);
				else
					angle = 2.0 * M_PI + atan(dz / dx);
			} else if (dx < 0) {
				if (dz >= 0)
					angle = M_PI + atan(dz / dx);
				else
					angle = M_PI + atan(dz / dx);
			} else {
				if (dz >= 0)
					angle = M_PI_2;
				else
					angle = M_PI + M_PI_2;
			}
			angle = angle * 180.0 / M_PI;	// passa in gradi
			// sistema l'angolo della camera in modo che sia
			// "confrontabile" con angle
			camera = gCamera.angle - 90.0;
			if (camera > 360.0)
				camera -= 360.0;
			else if (camera < 0)
				camera += 360.0;
			camera = 360.0 - camera;
			// calcola differenza e normalizza
			delta = fabsf(angle - camera);
			if (delta > 180.0)
				delta = 360.0 - delta;
			// dal delta possiamo capire se l'oggetto
			// cade nel nostro campo visivo...
			if (delta > FOV)
				goto nextObj;
		}
			
		glPushMatrix();
		// trasla e ruota il sistema di coordinate dell'oggetto
		// a seconda della sua posizione nello spazio
		glTranslatef(o->position.x, o->position.y, -o->position.z);
		if (!(o->flags & FLAGS_2D)) {
			glRotatef(o->pitch, 1.0, 0.0, 0.0);
			glRotatef(o->yaw, 0.0, 1.0, 0.0);
			glRotatef(o->roll, 0.0, 0.0, 1.0);
		} else {
			glRotatef(-gCamera.angle, 0.0, 1.0, 0.0);
			glRotatef(-gCamera.elevation, 1.0, 0.0, 0.0);
		}

		switch (o->model) {
			case MODEL_FIREBALL:
				{
					GLfloat lpos[] = {0, 0, 0, 1.0};
					GLfloat lcolor[] = {1.0, 0.95, 0.7, 1.0};

					glDisable(GL_TEXTURE_2D);
					glEnable(GL_BLEND);
					glDisable(GL_FOG);
					glDisable(GL_LIGHTING);
					glBlendFunc(GL_SRC_ALPHA, GL_DST_ALPHA);
					glColor4f(1.0, 1.0, 0.4, 0.4);
					gluSphere(notexquadr, o->radius, 8, 8);

					// attiva un punto luce omnidirezionale nel centro
					// della fireball, con attenuazione lineare rispetto
					// alla distanza. uuh.
					glDisable(GL_LIGHT1);
					glEnable(GL_LIGHTING);
					glLightfv(GL_LIGHT1, GL_POSITION, lpos);
					glLightfv(GL_LIGHT1, GL_DIFFUSE, lcolor);
					glLightf(GL_LIGHT1, GL_LINEAR_ATTENUATION, 0.3);
					glEnable(GL_LIGHT1);
				}
				break;
			case MODEL_SMOKEPUFF:
				glDisable(GL_TEXTURE_2D);
				glDepthMask(GL_FALSE);	// depth buffer read-only
				glDisable(GL_LIGHTING);
				if (gFogParams.on)	glEnable(GL_FOG);
				glEnable(GL_BLEND);
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
				{
					float	foo = 0.08 - o->decay * 0.08 + 0.02;
					float	bar = 0.707;	// sin(pi/4)
					
					glBegin(GL_TRIANGLE_FAN);
						glColor4f(0.6, 0.6, 0.65, o->decay * 0.7);
						glVertex3f(0, 0, 0);
						glColor4f(0.6, 0.6, 0.7, 0.0);
						glVertex3f(-foo * bar, -foo * bar, 0);
						glVertex3f(         0,       -foo, 0);
						glVertex3f( foo * bar, -foo * bar, 0);
						glVertex3f(       foo,          0, 0);
						glVertex3f( foo * bar,  foo * bar, 0);
						glVertex3f(         0,        foo, 0);
						glVertex3f(-foo * bar,  foo * bar, 0);
						glVertex3f(      -foo,          0, 0);
						glVertex3f(-foo * bar, -foo * bar, 0);
					glEnd();
				}
				glDepthMask(GL_TRUE);	// ripristina il depth buffer r/w
				break;
			case MODEL_BOUNCING:
				{
					float	height, r = o->radius;
					
					glEnable(GL_TEXTURE_2D);
					glDisable(GL_BLEND);
					glEnable(GL_LIGHTING);
					if (gFogParams.on)	glEnable(GL_FOG);
					glBindTexture(GL_TEXTURE_2D, tex_football);
					gluSphere(ballQuadr, r, 32, 32);
					height = o->position.y - gMap[o->mapSection].floorY;
					if (height >= 0.0 && height <= 1.0) {
						float	foo = 0.707;
						
						glRotatef(-o->roll, 0.0, 0.0, 1.0);
						glRotatef(-o->yaw, 0.0, 1.0, 0.0);
						glRotatef(-o->pitch, 1.0, 0.0, 0.0);
						glTranslatef(0, -height + 0.01, 0);
						glDisable(GL_TEXTURE_2D);
						glEnable(GL_BLEND);
						glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
						glDisable(GL_LIGHTING);
						glBegin(GL_TRIANGLE_FAN);
							glColor4f(0.0, 0.0, 0.0, 1.0 - height);
							glVertex3f( 0, 0,  0);
							glColor4f(0.0, 0.0, 0.0, 0.0);
							glVertex3f(-r*foo, 0, -r*foo);
							glVertex3f(     0, 0,     -r);
							glVertex3f( r*foo, 0, -r*foo);
							glVertex3f(     r, 0,      0);
							glVertex3f( r*foo, 0,  r*foo);
							glVertex3f(     0, 0,      r);
							glVertex3f(-r*foo, 0,  r*foo);
							glVertex3f(    -r, 0,      0);
							glVertex3f(-r*foo, 0, -r*foo);
						glEnd();
					}
				}
				break;
			case MODEL_EXPLOSION:
				{
					float	sphere_prec;
					GLfloat lpos[] = {0, 0, 0, 1};
					GLfloat lcolor[] = {o->decay, o->decay, o->decay * 0.9, 1.0};

					sphere_prec = (o->radius + 0.2) * 18.0;
					glEnable(GL_TEXTURE_2D);
					glBindTexture(GL_TEXTURE_2D, tex_flames);
					glEnable(GL_BLEND);
					glDisable(GL_FOG);
					glDisable(GL_LIGHTING);
					glBlendFunc(GL_SRC_ALPHA, GL_ONE);
					glColor4f(o->decay*0.6+0.4, o->decay*0.6+0.4, 0, o->decay);
					gluSphere(quadr, 0.5*sin((1.0 - o->decay)*2*M_PI), sphere_prec / 2, sphere_prec / 2);
					gluDisk(quadr, 0.3, o->radius*o->radius*2+0.1, sphere_prec, sphere_prec);
					
					glDisable(GL_LIGHT1);
					glEnable(GL_LIGHTING);
					glLightfv(GL_LIGHT1, GL_POSITION, lpos);
					glLightfv(GL_LIGHT1, GL_DIFFUSE, lcolor);
					glLightf(GL_LIGHT1, GL_LINEAR_ATTENUATION, 0.1);
					glEnable(GL_LIGHT1);
				}
				break;
			case MODEL_DOGE:
				glDisable(GL_LIGHTING);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, tex_doge);
				glDepthMask(GL_FALSE);
				glEnable(GL_BLEND);
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
				glNormal3f(0, 0, 1);
				glColor4f(1.0, 1.0, 1.0, 1.0);
				glBegin(GL_QUADS);
					glTexCoord2f(0,    1);	glVertex3f(-0.4, -0.43, 0);
					glTexCoord2f(1,    1);	glVertex3f( 0.4, -0.43, 0);
					glTexCoord2f(1, 0.06);	glVertex3f( 0.4,  0.43, 0);
					glTexCoord2f(0, 0.06);	glVertex3f(-0.4,  0.43, 0);
				glEnd();
				glDepthMask(GL_TRUE);
				glEnable(GL_LIGHTING);
				break;
			case MODEL_ANTINAZI:
				if (tex_antinazi) {
					glEnable(GL_TEXTURE_2D);
					glBindTexture(GL_TEXTURE_2D, tex_antinazi);
					glEnable(GL_LIGHTING);
					glEnable(GL_BLEND);
					if (gFogParams.on)	glEnable(GL_FOG);
					glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
					glColor4f(1.0, 1.0, 1.0, 1.0);
					glNormal3f(0, 0, -1);
					glBegin(GL_POLYGON);
						glTexCoord2f(1, 1);	glVertex3f(-0.1, -0.1, 0);
						glTexCoord2f(0, 1);	glVertex3f( 0.1, -0.1, 0);
						glTexCoord2f(0, 0);	glVertex3f( 0.1,  0.1, 0);
						glTexCoord2f(1, 0); glVertex3f(-0.1,  0.1, 0);
					glEnd();
				}
				break;
			case MODEL_PARTICLETORCH:
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, tex_gate_icon);
				glDisable(GL_BLEND);
				glEnable(GL_LIGHTING);
				if (gFogParams.on)	glEnable(GL_FOG);
				glNormal3f(0, 0, 1);
				glBegin(GL_QUADS);
					glTexCoord2f(0, 0);	glVertex3f(-0.05, 0.7, 0.02);
					glTexCoord2f(1, 0);	glVertex3f( 0.05, 0.7, 0.02);
					glTexCoord2f(1, 1);	glVertex3f( 0.05, 0.8, 0.02);
					glTexCoord2f(0, 1);	glVertex3f(-0.05, 0.8, 0.02);
				glEnd();

				glBindTexture(GL_TEXTURE_2D, envmap2);
				EnableTexEnvMap();
				glPushMatrix();
					glRotatef(-90.0, 1.0, 0.0, 0.0);
					gluCylinder(quadr, 0.02, 0.02, 1.0, 16, 4);
				glPopMatrix();
				glBindTexture(GL_TEXTURE_2D, textures[1]);
				DisableTexEnvMap();
				gluSphere(quadr, 0.1, 16, 16);
				
				glBindTexture(GL_TEXTURE_2D, textures[10]);
				glTranslatef(0.0, 0.5, 0.0);
				gluSphere(quadr, 0.03, 8, 8);
				gluCylinder(quadr, 0.02, 0.01, 0.05, 8, 1);
				glTranslatef(0.0, 0.0, 0.05);
				gluDisk(quadr, 0, 0.01, 8, 2);
				glTranslatef(0.0, 0.0, -0.01);
				glRotatef(90, 0.0, 1.0, 0.0);
				if ((o->other & SWITCH_ANIM_MASK) != SWITCH_ANIM_MASK) {
					float	tapAngle = ((o->other & SWITCH_ANIM_MASK) >> 24) * 90.0 / 70.0;
					
					if (!(o->other & SWITCH_ON_MASK))
						tapAngle = 90.0 - tapAngle;
					glRotatef(tapAngle, 1.0, 0.0, 0.0);
				} else {
					if (o->other & SWITCH_ON_MASK)
						glRotatef(90, 1.0, 0.0, 0.0);
				}
				gluCylinder(quadr, 0.008, 0.01, 0.1, 8, 2);
				break;
			case MODEL_TANK:
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, tex_kerosene);
				glDisable(GL_BLEND);
				glEnable(GL_LIGHTING);
				if (gFogParams.on)	glEnable(GL_FOG);
				gluCylinder(quadr, 0.25, 0.25, 0.7, 16, 2);
				break;
			case MODEL_HANDRAIL:
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, tex_geek);
				EnableTexEnvMap();
				if (gFogParams.on)	glEnable(GL_FOG);
				glDisable(GL_BLEND);
				glEnable(GL_LIGHTING);
				glPushMatrix();
					glRotatef(-90, 1.0, 0.0, 0.0);
					gluCylinder(quadr, 0.015, 0.015, 0.5, 8, 2);
				glPopMatrix();
				glTranslatef(0.0, 0.5, 0.0);
				gluSphere(quadr, 0.03, 10, 10);
				DisableTexEnvMap();
				// OMBRA PORTATA:
				// Proiezione sul pavimento del vettore luce (<sunlight> nell'rml)
				// che parte dalla sommita' del paletto.
				// FIXME: ottimizzare
				glTranslatef(0.0, -0.5, 0.0);
				glDisable(GL_TEXTURE_2D);
				{
					Polygon3D	*poly;
					Vector3D	pn, proj, palo = {0, 0.5, 0};
					float		cos45 = 0.03 * cos(M_PI/4.0), norma_q_rec;
					Point3D		p1, p2;	// vertici del poligono ombra

					// calcola la proiezione del paletto sul pavimento,
					// lungo la direzione del vettore luce
					poly = &gStaticPolys[gMap[o->mapSection].floorPolyID];
					pn = poly->normal;
					ProjectVector(&palo, &gGlobalLight.dir, &pn, &proj);
					// costruisci un quadrilatero lungo questa proiezione
					// e dello stesso spessore del paletto (ombra del paletto stesso)
					norma_q_rec = 0.015 / sqrt(proj.x * proj.x + proj.z * proj.z);
					p1.x = -proj.z * norma_q_rec;	p1.z = proj.x * norma_q_rec;
					p2.x = proj.z * norma_q_rec;	p2.z = -proj.x * norma_q_rec;
					// renderizza l'ombra
					glDisable(GL_TEXTURE_2D);
					glEnable(GL_BLEND);
					glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
					glDisable(GL_LIGHTING);
					glBegin(GL_QUADS);
						glColor4f(0.1, 0.1, 0.1, 0.6);
						glVertex3f(p1.x, 0, -p1.z);
						glVertex3f(p2.x, 0, -p2.z);
						glColor4f(0.1, 0.1, 0.1, 0.6);
						glVertex3f(proj.x + p2.x, 0, -proj.z - p2.z);
						glVertex3f(proj.x + p1.x, 0, -proj.z - p1.z);
					glEnd();
					// disco per simulare l'ombra della sfera in cima al paletto
					glBegin(GL_TRIANGLE_FAN);
						glVertex3f(proj.x,			0.001, -proj.z);
						glVertex3f(proj.x + 0.03,	0.001, -proj.z);
						glVertex3f(proj.x + cos45,	0.001, -proj.z - cos45);
						glVertex3f(proj.x,			0.001, -proj.z - 0.03);
						glVertex3f(proj.x - cos45,	0.001, -proj.z - cos45);
						glVertex3f(proj.x - 0.03,	0.001, -proj.z);
						glVertex3f(proj.x - cos45,	0.001, -proj.z + cos45);
						glVertex3f(proj.x,			0.001, -proj.z + 0.03);
						glVertex3f(proj.x + cos45,	0.001, -proj.z + cos45);
						glVertex3f(proj.x + 0.03,	0.001, -proj.z);
					glEnd();
				}
				break;
			case MODEL_BRICK:
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, textures[0]);
				glDisable(GL_BLEND);
				glEnable(GL_LIGHTING);
				if (gFogParams.on)	glEnable(GL_FOG);
				
				glBegin(GL_QUADS);
					glNormal3f(0, -1, 0);
					glTexCoord2f(0,    0);		glVertex3f(-0.1, -0.03, -0.05);
					glTexCoord2f(0.41, 0);		glVertex3f( 0.1, -0.03, -0.05);
					glTexCoord2f(0.41, 0.22);	glVertex3f( 0.1, -0.03,  0.05);
					glTexCoord2f(0,    0.22);	glVertex3f(-0.1, -0.03,  0.05);

					glNormal3f(0, 1, 0);
					glTexCoord2f(0,    0);		glVertex3f(-0.1, 0.03, -0.05);
					glTexCoord2f(0.41, 0);		glVertex3f( 0.1, 0.03, -0.05);
					glTexCoord2f(0.41, 0.22);	glVertex3f( 0.1, 0.03,  0.05);
					glTexCoord2f(0,    0.22);	glVertex3f(-0.1, 0.03,  0.05);

					glNormal3f(-1, 0, 0);
					glTexCoord2f(0.82, 0.23);	glVertex3f(-0.1, -0.03,  0.05);
					glTexCoord2f(0.82, 0.45);	glVertex3f(-0.1, -0.03, -0.05);
					glTexCoord2f(0.95, 0.45);	glVertex3f(-0.1,  0.03, -0.05);
					glTexCoord2f(0.95, 0.23);	glVertex3f(-0.1,  0.03,  0.05);

					glNormal3f(1, 0, 0);
					glTexCoord2f(0.82, 0.23);	glVertex3f( 0.1, -0.03,  0.05);
					glTexCoord2f(0.82, 0.45);	glVertex3f( 0.1, -0.03, -0.05);
					glTexCoord2f(0.95, 0.45);	glVertex3f( 0.1,  0.03, -0.05);
					glTexCoord2f(0.95, 0.23);	glVertex3f( 0.1,  0.03,  0.05);

					glNormal3f(0, 0, -1);
					glTexCoord2f(0.96, 0.36);	glVertex3f(-0.1, -0.03, -0.05);
					glTexCoord2f(1.34, 0.36);	glVertex3f( 0.1, -0.03, -0.05);
					glTexCoord2f(1.34, 0.23);	glVertex3f( 0.1,  0.03, -0.05);
					glTexCoord2f(0.96, 0.23);	glVertex3f(-0.1,  0.03, -0.05);
					
					glNormal3f(0, 0, 1);
					glTexCoord2f(0.96, 0.36);	glVertex3f(-0.1, -0.03, 0.05);
					glTexCoord2f(1.34, 0.36);	glVertex3f( 0.1, -0.03, 0.05);
					glTexCoord2f(1.34, 0.23);	glVertex3f( 0.1,  0.03, 0.05);
					glTexCoord2f(0.96, 0.23);	glVertex3f(-0.1,  0.03, 0.05);
				glEnd();
				break;
			case MODEL_FOOBAR:
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, envmap3);
				EnableTexEnvMap();
				glDisable(GL_BLEND);
				glEnable(GL_LIGHTING);
				if (gFogParams.on)	glEnable(GL_FOG);
				glColor4f(cos((o->pitch-180.0) * (M_PI/2) / 180.0) * 0.5 + 0.5,
						cos((o->roll-180.0) * (M_PI/2) / 180.0) * 0.5 + 0.5,
						cos((o->yaw-180.0) * (M_PI/2) / 180.0) * 0.5 + 0.5,
						1.0);
				gluCylinder(notexquadr, 0.1, 0.1, 0.2, 10, 10);
				gluCylinder(notexquadr, 0.3, 0.3, 0.2, 16, 16);
				glTranslatef(0, 0, 0.0);
				gluDisk(notexquadr, 0.1, 0.3, 16, 20);
				glTranslatef(0, 0, 0.2);
				gluDisk(notexquadr, 0.1, 0.3, 16, 20);
				DisableTexEnvMap();
				break;
			case MODEL_LAMP:
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, envmap3);
				glEnable(GL_LIGHTING);
				{
					GLint	lid = (GLint)o->other;
					GLfloat	diff[] = {1.0, 1.0, 1.0, 1.0},
							pos[] = {0.0, 0.0, 0.0, 1.0},
							dir[] = {0.0, -1.0, 0.0};

					glDisable(lid);
					glLightfv(lid, GL_POSITION, pos);			// posizione
					glLightfv(lid, GL_SPOT_DIRECTION, dir);		// direzione
					glLightf(lid, GL_SPOT_CUTOFF, 43.0);		// ampiezza del cono luce
					glLightf(lid, GL_SPOT_EXPONENT, 1.0);		// "sfocatura" del cono luce
					glLightfv(lid, GL_DIFFUSE, diff);
					glLightf(lid, GL_LINEAR_ATTENUATION, 0.3);
					glEnable(lid);
				}
				if (gFogParams.on)	glEnable(GL_FOG);
				EnableTexEnvMap();
				glBegin(GL_QUADS);
					glNormal3f(0, -1, 0);
					glVertex3f(-0.1, -0.1, -0.1);
					glVertex3f( 0.1, -0.1, -0.1);
					glVertex3f( 0.1, -0.1,  0.1);
					glVertex3f(-0.1, -0.1,  0.1);
					glNormal3f(0, 1, 0);
					glVertex3f(-0.1, 0.1, -0.1);
					glVertex3f( 0.1, 0.1, -0.1);
					glVertex3f( 0.1, 0.1,  0.1);
					glVertex3f(-0.1, 0.1,  0.1);
					glNormal3f(-1, 0, 0);
					glVertex3f(-0.1, -0.1,  0.1);
					glVertex3f(-0.1, -0.1, -0.1);
					glVertex3f(-0.1,  0.1, -0.1);
					glVertex3f(-0.1,  0.1,  0.1);
					glNormal3f(1, 0, 0);
					glVertex3f( 0.1, -0.1,  0.1);
					glVertex3f( 0.1, -0.1, -0.1);
					glVertex3f( 0.1,  0.1, -0.1);
					glVertex3f( 0.1,  0.1,  0.1);
					glNormal3f(0, 0, -1);
					glVertex3f(-0.1, -0.1, -0.1);
					glVertex3f( 0.1, -0.1, -0.1);
					glVertex3f( 0.1,  0.1, -0.1);
					glVertex3f(-0.1,  0.1, -0.1);
					glNormal3f(0, 0, 1);
					glVertex3f(-0.1, -0.1, 0.1);
					glVertex3f( 0.1, -0.1, 0.1);
					glVertex3f( 0.1,  0.1, 0.1);
					glVertex3f(-0.1,  0.1, 0.1);
				glEnd();
				glDisable(GL_TEXTURE_2D);
				glDisable(GL_LIGHTING);
				glColor4f(1.0, 1.0, 1.0, 1.0);
				glNormal3f(0, -1, 0);
				glBegin(GL_QUADS);
					glVertex3f(-0.05, -0.101, -0.05);
					glVertex3f( 0.05, -0.101, -0.05);
					glVertex3f( 0.05, -0.101,  0.05);
					glVertex3f(-0.05, -0.101,  0.05);
				glEnd();
				DisableTexEnvMap();
				break;
			case MODEL_GRASS:	// erbetta
				glDisable(GL_LIGHTING);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, tex_piantine[o->other]);
				glDepthMask(GL_FALSE);
				glEnable(GL_BLEND);
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
				glNormal3f(0, 0, 1);
				glColor4f(1.0, 1.0, 1.0, 1.0);
				glBegin(GL_QUADS);
					glTexCoord2f(0, 1);  glVertex3f(-0.2, -0.2, 0);
					glTexCoord2f(1, 1);  glVertex3f( 0.2, -0.2, 0);
					glTexCoord2f(1, 0);  glVertex3f( 0.2,  0.2, 0);
					glTexCoord2f(0, 0);  glVertex3f(-0.2,  0.2, 0);
				glEnd();
				glDepthMask(GL_TRUE);
				glEnable(GL_LIGHTING);
				break;
			default:
				break;
		}
		glPopMatrix();
		gDrawnObjects++;
nextObj:
		o = o->next;
	}
}

// controlla se il giocatore ha sbattuto contro un oggetto,
// e in tal caso modifica il vettore step in modo da deviarlo
void CheckPlayerObjectCollision(Point3D pos, Vector3D *step)
{
	float		dist, dx, dz, dot;
	Object		*o = ObjectClosestToPoint(pos, NULL);
	Vector3D	normal, up, new_step;

	if (o == NULL)
		return;	// nessun oggetto
	if (!(o->flags & SOLID))
		return;
	// innanzi tutto controlla la distanza giocatore-oggetto
	// lungo la direzione verticale...
	if ((o->position.y - gCamera.y) > o->radius)
		return;	// oggetto troppo in alto
	if (o->position.y < (gCamera.y - gCamera.height) && (gCamera.y - gCamera.height - o->position.y) > o->radius)
		return;	// oggetto troppo in basso
	
	// quindi calcola la distanza tra l'asse del giocatore
	// e la posizione dell'oggetto
	dx = o->position.x - gCamera.x;
	dz = o->position.z - gCamera.z;
	dist = dx * dx + dz * dz;
	if (dist > (o->radius + gCamera.radius)*(o->radius+gCamera.radius))
		return;	// troppo distanti
	
	// ora sappiamo che l'oggetto "interseca" il cilindro del giocatore
	// e dobbiamo far "scivolare" il giocatore contro l'oggetto.
	// calcola il vettore normale alla sfera che contiene l'oggetto
	// nel punto colpito dal giocatore
	normal.x = dx;
	normal.y = 0;
	normal.z = dz;
	Normalize(&normal);
	dot = (normal.x * step->x) + (normal.z * step->z);
	// controlla se ci stiamo allontanando dall'oggetto,
	// in questo caso lascia lo step com'e' dato  che
	// ci stiamo allontanando dall'oggetto
	if (dot < 0)
		return;
	// se ci stiamo scontrando con l'oggetto,
	// con due prodotti vettoriali otteniamo la nuova velocita'
	// e la rallentiamo leggermente per rendere l'idea dell'attrito
	// con l'oggetto
	CrossProduct(normal, *step, &up);
	CrossProduct(normal, up, &new_step);
	step->x = -new_step.x * 0.8;
	step->y = -new_step.y * 0.8;
	step->z = -new_step.z * 0.8;
}

// ritorna un puntatore all'oggetto piu' vicino ad un punto
// e il quadrato della distanza
Object *ObjectClosestToPoint(Point3D p, float *dist)
{
	Object  *o = gFirstObject, *found = NULL;
	float	min_dist = -1;

	// passa ogni oggetto...
	while (o != NULL) {
		float	dx, dy, dz, d;

		if (o->model == MODEL_SMOKEPUFF)
			goto skipThis;

		// calcola la distanza tra l'oggetto
		// e il nostro punto... che bello, non
		// serve nemmeno fare la radice quadrata ;-)
		dx = p.x - o->position.x;
		dy = p.y - o->position.y;
		dz = p.z - o->position.z;
		d = dx * dx + dy * dy + dz * dz;
		if (d < min_dist || min_dist == -1) {
			min_dist = d;
			found = o;
		}
skipThis:
		o = o->next;
	}
	if (dist != NULL)
		*dist = min_dist;
	return found;
}

int GetModelIDFromName(char *n)
{
	if (strcasecmp(n, "Doge") == 0)
		return MODEL_DOGE;
	else if (strcasecmp(n, "AntiNazi") == 0)
		return MODEL_ANTINAZI;
	else if (strcasecmp(n, "BouncingBall") == 0)
		return MODEL_BOUNCING;
	else if (strcasecmp(n, "ParticleTorch") == 0)
		return MODEL_PARTICLETORCH;
	else if (strcasecmp(n, "Tank") == 0)
		return MODEL_TANK;
	else if (strcasecmp(n, "Handrail") == 0)
		return MODEL_HANDRAIL;
	else if (strcasecmp(n, "Brick") == 0)
		return MODEL_BRICK;
	else if (strcasecmp(n, "Foobar") == 0)
		return MODEL_FOOBAR;
	else if (strcasecmp(n, "Lamp") == 0)
		return MODEL_LAMP;
	else if (strcasecmp(n, "Grass") == 0)
		return MODEL_GRASS;
	else
		return NONE;
}

