typedef struct Object {         // struttura basilare per qualsiasi oggetto
	struct Object   *prev;
	struct Object   *next;
	
	Point3D		position;
	Vector3D	speed;
	float		pitch;	// rotaz asse x
	float		yaw;	// rotaz asse y
	float		roll;	// rotaz asse z
	
	int		mapSection;
	int		model;
	int		collDetDone;	// collision detection gia' fatto
	unsigned long	flags;
	unsigned long	other;	// dati vari dipendenti dal tipo di oggetto
	float		radius;
	float		decay;  // valore usato come timer per far svanire l'oggetto
	void		(*destroyProc)(struct Object *o);
} Object;

// model
enum {
	MODEL_FIREBALL = 1,
	MODEL_SMOKEPUFF,
	MODEL_EXPLOSION,
	MODEL_DOGE,
	MODEL_PARTICLETORCH,
	MODEL_ANTINAZI,
	MODEL_BOUNCING,
	MODEL_TANK,
	MODEL_HANDRAIL,
	MODEL_BRICK,
	MODEL_FOOBAR,
	MODEL_LAMP,
	MODEL_GRASS
};

#define OBJECT_ACTION_MAX_DIST	2

// flags
#define FLAGS_2D	0x0001
#define	IS_SWITCH	0x0002
#define SOLID		0x0004
#define FIXED		0x0008

// altri dati
#define SWITCH_ON_MASK		0x8000
#define SWITCH_ANIM_MASK	0xff000000

void InitObjects(void);
void NewObject(int model, Point3D start_pos, Vector3D start_speed, float ax, float ay, float az);
void DestroyObject(Object *o);
void UpdateObjects(double dt);
void UpdateObjectPosition(Object *obj, double deltaTime);	// obsoleta

void BroadcastObjectAction(void);

void RenderObjects(void);

void CheckPlayerObjectCollision(Point3D pos, Vector3D *step);
Object *ObjectClosestToPoint(Point3D p, float *dist);

int GetModelIDFromName(char *n);

