/********************************************************************************
Rockland
1999-oo Tito Dal Canton
	Funzioni di rendering di mondo, oggetti ed effettini vari.

TODO: prima o poi bisogna implementare la suddivisione dei poligoni grandi
      in "pezzetti" (tasselli) piu' piccoli, perche' cosi' facendo l'illuminazione
	  basata sui vertici funziona MOLTO meglio. E magari riuscirei anche ad
	  implementare le ombre portate... migliorerebbero molto la cosa.
********************************************************************************/
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <SDL.h>
#include <SDL_opengl.h>
#include "global.h"
#include "worlddata.h"
#include "objects.h"
#include "font-engine.h"
#include "textures.h"
#include "3d-calc.h"
#include "render.h"

#define OPAQUE_POLYS		0
#define TRANSPARENT_POLYS	1

static void RenderPolys(int poly_type);
static void RenderShadows(void);
static void RenderSky(void);
static void ActivateSceneGeometry(void);
static void ActivateScreenGeometry(void);
static int HaveGLExt(char *name);

// estensioni OpenGL richiamate se disponibili
#ifndef GL_ARB_multitexture
#ifndef PFNGLMULTITEXCOORD2FARBPROC
typedef void (APIENTRY * PFNGLMULTITEXCOORD2FARBPROC)(GLenum target, GLfloat s, GLfloat t);
typedef void (APIENTRY * PFNGLACTIVETEXTUREARBPROC)(GLenum texture);
typedef void (APIENTRY * PFNGLCLIENTACTIVETEXTUREARBPROC)(GLenum texture);
#endif
PFNGLMULTITEXCOORD2FARBPROC		glMultiTexCoord2fARB = 0;
PFNGLACTIVETEXTUREARBPROC		glActiveTextureARB = 0;
PFNGLCLIENTACTIVETEXTUREARBPROC	glClientActiveTextureARB = 0;
#endif

extern	Camera			gCamera;
extern	Polygon3D		gStaticPolys[];
extern	long			gStaticPolysCount;
extern	GLuint			textures[];
extern	GLuint			envmap2;
		GLuint			gSkyTexture = 0;
static	GLuint			gCurrentTexture;
		int				gHaveMultitexture = 0;
static	int				gRenderedPolys;
extern	int				gScreenWidth, gScreenHeight;
extern	unsigned long	gDisplayFlags;
		GLfloat			gTeleportEffectDecay;
extern	int				currMapSection, gMapSectionsNum;
extern	float			fps;
		GLfloat			clearColor[3] = {-1.0, -1.0, -1.0};
		FogParams		gFogParams;
		LightSource		gGlobalLight;

// routine principale di rendering
void RocklandRender(void)
{
	GLbitfield	toClear = GL_DEPTH_BUFFER_BIT;
	
	if (clearColor[0] != -1.0)
		toClear |= GL_COLOR_BUFFER_BIT;
	if (gDisplayFlags & SHADOWS)
		toClear |= GL_STENCIL_BUFFER_BIT;
	glClear(toClear);

	ActivateSceneGeometry();
	glRotatef(gCamera.elevation, 1.0, 0.0, 0.0);
	glRotatef(gCamera.angle, 0.0, 1.0, 0.0);
	glTranslatef(-gCamera.x, -gCamera.y, gCamera.z);

	// luce del sole
	{
		GLfloat	d[4];

		d[0] = -gGlobalLight.dir.x;	d[1] = -gGlobalLight.dir.y;	d[2] = gGlobalLight.dir.z; d[3] = 0.0;
		glLightfv(GL_LIGHT0, GL_DIFFUSE, gGlobalLight.diffuse);
		glLightfv(GL_LIGHT0, GL_AMBIENT, gGlobalLight.ambient);
		glLightfv(GL_LIGHT0, GL_POSITION, d);
		glEnable(GL_LIGHT0);
	}
	
	gRenderedPolys = 0;
	// renderizza la scena
	RenderSky();
	RenderPolys(OPAQUE_POLYS);
	RenderPolys(TRANSPARENT_POLYS);
	RenderObjects();
	
	// renderizza cio' che non fa parte della scena
	ActivateScreenGeometry();
	// renderizza il fade del teletrasporto se e' il caso
	if (gTeleportEffectDecay > 0.0) {
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
		glDisable(GL_TEXTURE_2D);
		glColor4f(	(random() / RAND_MAX) + 1.0,
					(random() / RAND_MAX) + 1.0,
					(random() / RAND_MAX) + 1.0,
					gTeleportEffectDecay);
		glBegin(GL_TRIANGLE_FAN);
			glVertex3f(0, 0, 0);
			glVertex3f(gScreenWidth-1, 0, 0);
			glVertex3f(gScreenWidth-1, gScreenHeight-1, 0);
			glVertex3f(0, gScreenHeight-1, 0);
		glEnd();
	}
	
	// ora disegnamo le info in tempo reale (fps ecc.)
	if (gDisplayFlags & SHOW_INFO) {
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_SRC_ALPHA);
		glColor4f(0.0, 0.05, 0.0, 0.6);
		glDisable(GL_TEXTURE_2D);
		glBegin(GL_TRIANGLE_FAN);
			glVertex3f(0.0, 0.0, 0.0); glVertex3f(gScreenWidth-1, 0.0, 0.0);
			glVertex3f(gScreenWidth-1, 35.0, 0.0); glVertex3f(0.0, 35.0, 0.0);
		glEnd();
	
		glEnable(GL_TEXTURE_2D);
		glColor4f(1.0, 1.0, 1.0, 1.0);

		{
			char    str[256] = "";
		
			sprintf(str, "[%.1f FPS] [%d map_sections, %ld poligoni, %ld%% disegnati]",
							fps, gMapSectionsNum, gStaticPolysCount, (gRenderedPolys * 100 / gStaticPolysCount));
			font_drawstring(str, 1, 1);
			sprintf(str, "[X=%.2f Y=%.2f Z=%.2f] [azim=%.1f elev=%.1f] [map_sect=%d]",
						gCamera.x, gCamera.y, gCamera.z, gCamera.angle, gCamera.elevation, currMapSection);
			font_drawstring(str, 1, 17);
		}
	}
	
	SDL_GL_SwapBuffers();
}

static void RenderPolys(int poly_type)
{
	int		i, j;
	float	camX = gCamera.x, camY = gCamera.y, camZ = gCamera.z;

	// imposta tutti i parametri necessari
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);
	if (gHaveMultitexture)
		glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, textures[0]);
	if (gFogParams.on)	glEnable(GL_FOG);
	if (poly_type == TRANSPARENT_POLYS) {
		glDisable(GL_LIGHTING);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		// imposta il depth buffer read-only:
		// in questo modo i poligoni trasparenti, che stiamo
		// per disegnare, verranno correttamente "clippati"
		// in base a quanto abbiamo disegnato prima; pero'
		// non influenzeranno il depth buffer (essendo ora
		// read-only), e quindi non nasconderanno eventuali
		// poligoni che ci stanno dietro...
		glDepthMask(GL_FALSE);
	} else {
		glDisable(GL_LIGHTING);
		glDisable(GL_BLEND);
		glDepthMask(GL_TRUE);	// depth buffer normale
	}

	gCurrentTexture = 0;	
	for (i = 0; i < gStaticPolysCount; i++) {
		Polygon3D		poly = gStaticPolys[i];
		unsigned int	vertex = 0;
		int				texture = poly.texture;
		
		if (texture < 0 || texture >= kNumTextures)
			continue;
		if ((poly_type == TRANSPARENT_POLYS && poly.a == 1.0) ||
			(poly_type != TRANSPARENT_POLYS && poly.a != 1.0))
			continue;

		if (poly_type == OPAQUE_POLYS) {
			// se un poligono opaco ci volta le spalle,
			// non ci preoccupiamo nemmeno di considerarlo
			float   dot = poly.normal.x*camX + poly.normal.y*camY + poly.normal.z*camZ;

			if (dot < poly.dist)
				continue;
		}

		// cambia la texture se necessario
		if (texture != gCurrentTexture) {
			glBindTexture(GL_TEXTURE_2D, textures[texture]);
			gCurrentTexture = texture;
		}
		// multitexturing
		if (gHaveMultitexture) {
			glActiveTextureARB(GL_TEXTURE1_ARB);
			if (texture == 10 || texture == 1) {
				// effetto metallico semi-riflettente
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, envmap2);
				EnableTexEnvMap();
				glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			} else {
				// lightmap (poi andra' fatta ovviamente anche sul metallo...)
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, poly.lmap);
				DisableTexEnvMap();
				glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			}
			glActiveTextureARB(GL_TEXTURE0_ARB);
		}
		glColor4f(poly.r, poly.g, poly.b, poly.a);
		
		// passa i dati del poligono a OpenGL
		glBegin(GL_TRIANGLE_FAN);
		for (vertex = 0; vertex < poly.numVertices; vertex++) {
			float	px = poly.vertices[vertex].x,
					py = poly.vertices[vertex].y,
					pz = -poly.vertices[vertex].z,
					nx = poly.vertices[vertex].nx,
					ny = poly.vertices[vertex].ny,
					nz = -poly.vertices[vertex].nz,
					u = poly.vertices[vertex].u,
					v = -poly.vertices[vertex].v,
					lmu = poly.vertices[vertex].lm_u,
					lmv = poly.vertices[vertex].lm_v;

			glNormal3f(nx, ny, nz);
			if (gHaveMultitexture) {
				glMultiTexCoord2fARB(GL_TEXTURE0_ARB, u, v);
				glMultiTexCoord2fARB(GL_TEXTURE1_ARB, lmu, lmv);
			} else
				glTexCoord2f(u, v);
			glVertex3f(px, py, pz);
		}
		glEnd();

		// debug: renderizza le normali ai vertici
		if (gDisplayFlags & SHOW_NORMALS) {
			if (gFogParams.on)      glDisable(GL_FOG);
			glDisable(GL_LIGHTING);
			if (gHaveMultitexture && (gCurrentTexture == 10 || gCurrentTexture == 1)) {
				glActiveTextureARB(GL_TEXTURE1_ARB);
				glDisable(GL_TEXTURE_2D);
				glActiveTextureARB(GL_TEXTURE0_ARB);
				glDisable(GL_TEXTURE_2D);
			} else {
				glDisable(GL_TEXTURE_2D);
			}
			if (poly_type != TRANSPARENT_POLYS) {
				glEnable(GL_BLEND);
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
				glDepthMask(GL_FALSE);
			}
			for (vertex = 0; vertex < poly.numVertices; vertex++) {
				glColor4f(1.0, 0.3, 0.3, 1.0);	// normale al vertice
				glBegin(GL_LINES);
				glVertex3f(poly.vertices[vertex].x, poly.vertices[vertex].y, -poly.vertices[vertex].z);
				glVertex3f(poly.vertices[vertex].x + poly.vertices[vertex].nx * (1.0/7),
							poly.vertices[vertex].y + poly.vertices[vertex].ny * (1.0/7),
							-poly.vertices[vertex].z - poly.vertices[vertex].nz * (1.0/7));
				
				glColor4f(1.0, 1.0, 0.3, 1.0);	// normale al piano
				glVertex3f(poly.vertices[vertex].x, poly.vertices[vertex].y, -poly.vertices[vertex].z);
				glVertex3f(poly.vertices[vertex].x + poly.normal.x * (1.0/7),
							poly.vertices[vertex].y + poly.normal.y * (1.0/7),
							-poly.vertices[vertex].z - poly.normal.z * (1.0/7));

#if 1
				glColor4f(1.0, 1.0, 1.0, 1.0);	// vettore luce
				glVertex3f(poly.vertices[vertex].x, poly.vertices[vertex].y, -poly.vertices[vertex].z);
				glVertex3f(poly.vertices[vertex].x + gGlobalLight.dir.x,
								poly.vertices[vertex].y + gGlobalLight.dir.y,
								-poly.vertices[vertex].z - gGlobalLight.dir.z);
#endif
				glEnd();
			}
			if (poly_type != TRANSPARENT_POLYS) {
				glDisable(GL_BLEND);
				glDepthMask(GL_TRUE);
			}
			glDisable(GL_LIGHTING);
			glEnable(GL_TEXTURE_2D);
			if (gFogParams.on)	glEnable(GL_FOG);
		}
		gRenderedPolys++;
	}
	if (poly_type == TRANSPARENT_POLYS)
		glDepthMask(GL_TRUE);
	if (gHaveMultitexture) {
		glActiveTextureARB(GL_TEXTURE1_ARB);
		glDisable(GL_TEXTURE_2D);
		DisableTexEnvMap();
		glActiveTextureARB(GL_TEXTURE0_ARB);
	}
}

static void RenderShadows(void)
{
/*	unsigned int	i;
	float			camX = gCamera.x, camY = gCamera.y, camZ = gCamera.z;
	
	glColorMask(0, 0, 0, 0);
	glEnable(GL_STENCIL_TEST);
	glStencilFunc(GL_ALWAYS, 1, 1);				// passa sempre lo stencil test
	glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);	// poni a 1 dove disegni
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);

	for (i = 0; i < gStaticPolysCount; i++) {
		Polygon3D		poly = gStaticPolys[i];
		unsigned int	j;
		float			dot = poly.normal.x*camX + poly.normal.y*camY + poly.normal.z*camZ;

		if (dot < poly.dist || poly.texture == NONE)
			continue;
		for (j = 0; j < poly.numShadows; j++) {
			ShadowPoly		s = poly.shadows[j];
			unsigned int	vertex;
			
			glBegin(GL_POLYGON);
			for (vertex = 0; vertex < s.numVertices; vertex++) {
				float	px = s.vertices[vertex].x,
						py = s.vertices[vertex].y,
						pz = -s.vertices[vertex].z;

				glVertex3f(px, py, pz);
			}
			glEnd();
		}
	}

	glColorMask(1, 1, 1, 1);
	glStencilFunc(GL_EQUAL, 1, 1);				// disegna solo dove lo stencil vale 1
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);		// non modificarlo quando disegni
	
	// disegna un rettangolo che copra tutta la scena;
	// siccome abbiamo lo stencil, verranno pero' mostrate
	// solo le zone coperte dalle ombre ;-)
	glDisable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_FOG);
	glDisable(GL_LIGHTING);
	glPushMatrix();
	glLoadIdentity();
	glColor4f(0, 0, 0, 0.5);
	glBegin(GL_TRIANGLE_STRIP);
		glVertex3f(-0.1f,  0.1f, -0.1f);
		glVertex3f(-0.1f, -0.1f, -0.1f);
		glVertex3f( 0.1f,  0.1f, -0.1f);
		glVertex3f( 0.1f, -0.1f, -0.1f);
	glEnd();
	glPopMatrix();
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_STENCIL_TEST);*/
}

static void RenderSky(void)
{
	if (gSkyTexture == 0)
		return;
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, gSkyTexture);
	glDisable(GL_BLEND);
	glDisable(GL_LIGHTING);
	glDisable(GL_FOG);

	glColor4f(1.0, 1.0, 1.0, 1.0);
	glBegin(GL_TRIANGLE_FAN);
		glTexCoord2f(0, 0);	glVertex3f(-200, 9, -200);
		glTexCoord2f(1, 0);	glVertex3f( 200, 9, -200);
		glTexCoord2f(1, 1);	glVertex3f( 200, 9,  200);
		glTexCoord2f(0, 1);	glVertex3f(-200, 9,  200);
	glEnd();
}

// sequenza pre-mondo (teletrasporto, piu' un esperimento
// per effetti strani che una parte di rockland :)
void TeleportEffect(void)
{
	unsigned long	prevTime, i, duration = 1000;
	float			rects[10];
	
	ActivateScreenGeometry();
	
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	
	for (i = 0; i < 10; i++)
		rects[i] = (random() / (float)RAND_MAX) * gScreenWidth;
	
	prevTime = SDL_GetTicks();
	do {
		float	foo = sin((SDL_GetTicks() - prevTime) * 2 * M_PI / duration) * gScreenWidth/2.0;
		
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glColor4f(0.0, 0.0, 0.0, (SDL_GetTicks() - prevTime) / (float)duration);
		glBegin(GL_TRIANGLE_STRIP);
		glVertex3f(0, gScreenHeight, 0);
		glVertex3f(0, 0, 0);
		glVertex3f(gScreenWidth, gScreenHeight, 0);
		glVertex3f(gScreenWidth, 0, 0);
		glEnd();
		glBlendFunc(GL_SRC_ALPHA, GL_DST_ALPHA);
		for (i = 0; i < 100; i++) {
			float	x = (random() / (float)RAND_MAX) * gScreenWidth,
					y = (random() / (float)RAND_MAX) * gScreenHeight,
					x1, y1;

			x1 = 50;
			y1 = 50;
			glBegin(GL_TRIANGLE_FAN);
			glColor4f(0, 0.5, 0.8, 1.0);
			glVertex3f(x, y, 0);
			glColor4f(0, 0.5, 0.8, 0.0);
			glVertex3f(x + x1, y + y1, 0);
			glVertex3f(x + x1, y - y1, 0);
			glVertex3f(x - x1, y - y1, 0);
			glVertex3f(x - x1, y + y1, 0);
			glEnd();
		}
		for (i = 0; i < 10; i++) {
//			rects[i] *= 0.99;
			glColor4f(0.3, 0.9, 1.0, 0.5);
			glBegin(GL_QUADS);
			glVertex3f(rects[i]+foo, 0, 0);
			glVertex3f(rects[i]*2+foo, 0, 0);
			glVertex3f(rects[i]*2+foo, gScreenHeight, 0);
			glVertex3f(rects[i]+foo, gScreenHeight, 0);
			glEnd();
		}
		SDL_GL_SwapBuffers();
	} while ((SDL_GetTicks() - prevTime) < duration);
}

void InitOpenGL(void)
{
	printf("Motore OpenGL: %s\n", glGetString(GL_RENDERER));

	glShadeModel(GL_SMOOTH);
	glLineWidth(1.5);
	glPolygonMode(GL_FRONT, GL_FILL);
	glDepthFunc(GL_LESS);
	
	// colore di sfondo
	if (clearColor[0] != -1)
		glClearColor(clearColor[0], clearColor[1], clearColor[2], 1.0);
	
	// nebbia
	if (gFogParams.on && gFogParams.dist > 0.0) {
		GLfloat	temp[3] = {gFogParams.color[0], gFogParams.color[1], gFogParams.color[2]};
		GLfloat	dens = 1.0 / gFogParams.dist;
		
		glEnable(GL_FOG);
		glFogfv(GL_FOG_COLOR, temp);
		glFogf(GL_FOG_DENSITY, dens);
		glHint(GL_FOG_HINT, GL_NICEST);
	}

	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	if (gDisplayFlags & SHOW_NORMALS) {
		glEnable(GL_LINE_SMOOTH);
		glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
	}

	// richiama le estensioni se disponibili
	if (HaveGLExt("GL_ARB_multitexture")) {
#ifndef GL_ARB_multitexture
		glMultiTexCoord2fARB = (PFNGLMULTITEXCOORD2FARBPROC)glXGetProcAddressARB("glMultiTexCoord2fARB");
		glActiveTextureARB = (PFNGLACTIVETEXTUREARBPROC)glXGetProcAddressARB("glActiveTextureARB");
		glClientActiveTextureARB = (PFNGLCLIENTACTIVETEXTUREARBPROC)glXGetProcAddressARB("glClientActiveTextureARB");
#endif
		gHaveMultitexture = 1;
		fprintf(stderr, "GL_ARB_multitexture presente. Evviva.\n");
	}
}

void QuitOpenGL(void)
{
	// rimetti le cose a posto, altrimenti SDL
	// tende a non capire piu' niente :-P
	glDisable(GL_FOG);
	glDisable(GL_LIGHTING);
	glDisable(GL_BLEND);
	glDisable(GL_STENCIL_TEST);
	glColorMask(1, 1, 1, 1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

// attiva le matrici di proiezione prospettica:
// prepara il motore a renderizzare la scena 3D
static void ActivateSceneGeometry(void)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(FOV, (float)gScreenWidth / (float)gScreenHeight, 0.01f, 1000.0f);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

// attiva le matrici di proiezione ortogonale su schermo
static void ActivateScreenGeometry(void)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0, (float)(gScreenWidth-1), (float)(gScreenHeight-1), 0.0, -1.0, 1.0);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	if (gFogParams.on)	glDisable(GL_FOG);
}

// controlla se questa macchina dispone
// di una certa estensione OpenGL.
static int HaveGLExt(char *name)
{
	char	*s = (char *)glGetString(GL_EXTENSIONS);

	if (strstr(s, name) != NULL)
		return 1;
	else
		return 0;
}

