/*****************************************************************************
Rockland
2001-oo Tito Dal Canton
	Caricamento delle texture da file
*****************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <errno.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <SDL.h>
#include <SDL_opengl.h>
#include <SDL_image.h>	/* possiamo caricare di tutto (Morpheus, The Matrix) */
#include "global.h"
#include "3d-calc.h"
#include "textures.h"

/* lato delle texture usate come lightmap */
#define LIGHTMAP_RES	256

extern LightSource	gGlobalLight;

/* carica un'immagine e ricavane una texture per OpenGL */
GLuint LoadTexture(char *filename, int alphaMode, GLint filter)
{
	SDL_Surface	*bmp, *image;
	GLuint		newTex;
	SDL_Rect	r = {0, 0, 0, 0};
	char		fne[1024] = "";
	int			h, w, x, y;

	/* prova ad aprire il file (con varie possibili estensioni) e
	   ricava una surface SDL (UNA RIGA di codice, grazie a SDL_image) */
	bmp = IMG_Load(filename);
	if (bmp == NULL) {
		sprintf(fne, "%s.bmp", filename);
		bmp = IMG_Load(fne);
		if (bmp == NULL) {
			sprintf(fne, "%s.jpg", filename);
			bmp = IMG_Load(fne);
			if (bmp == NULL) {
				sprintf(fne, "%s.gif", filename);
				bmp = IMG_Load(fne);
				if (bmp == NULL) {
					fprintf(stderr, "%s[bmp,jpg,gif]: %s\n", filename, SDL_GetError());
					return 0;
				}
			}
		}
	}
	h = bmp->h;	w = bmp->w;
	
	/* siccome la surface appena ricavata non sara' probabilmente nel
	   formato richiesto, ne creiamo una nuova con le caratteristiche
	   che ci servono, e poi ci copiamo la prima */
	image = SDL_CreateRGBSurface(
				SDL_SWSURFACE,
				w, h,
				32,
#if SDL_BYTEORDER == SDL_LIL_ENDIAN	/* attenzione al big/little endian per le mask RGBA */
				0x000000FF,	0x0000FF00,	0x00FF0000,	0xFF000000
#else
				0xFF000000,	0x00FF0000,	0x0000FF00,	0x000000FF
#endif
			);
	if (image == NULL) {
		fprintf(stderr, "%s: %s\n", filename, SDL_GetError());
		return 0;
	}
	/* copia una surface dentro l'altra, ignorando il canale alpha originale */
	SDL_SetAlpha(bmp, 0, 0);
	r.w = w;	r.h = h;
	SDL_BlitSurface(bmp, &r, image, &r);
	
	/* calcola il canale alfa a seconda di cosa dice alphaMode */
	SDL_LockSurface(image);
	for (y = 0; y < h; y++) {
		for (x = 0; x < w; x++) {
			unsigned long	*pixel = ((unsigned long *)image->pixels) + x + y * w;
			unsigned char	*r, *g, *b, *a;

			// TODO: usare i valori mask, shift e loss in SDL_PixelFormat per ricavare
			//       R G e B, invece di questo sistema pericoloso e poco portabile.
			r = ((unsigned char *)pixel) + 0;
			g = ((unsigned char *)pixel) + 1;
			b = ((unsigned char *)pixel) + 2;
			a = (unsigned char *)pixel + 3;
			*a = 255;	/* opaco per default */
			switch (alphaMode) {
				case ALPHAMODE_RED:
					if (*r == 255 && *g == 0 && *b == 0)
						*a = 0;
					break;
				case ALPHAMODE_GREEN:
					if (*r <= 2 && *g >= 240 && *b <= 2) {
						*a = 0;
						*g = 0;
					}
					break;
				case ALPHAMODE_BLUE:
					*a = 255 - *b;
					*b = 0;
					break;
				case ALPHAMODE_COLOR_AVERAGE:
					*a = (*r + *b + *g) / 3;
					break;
				default:
					break;
			}
		}
	}
	
	/* crea una nuova texture OpenGL */
	glGenTextures(1, &newTex);
	glBindTexture(GL_TEXTURE_2D, newTex);
	gluBuild2DMipmaps(GL_TEXTURE_2D, 4, w, h, GL_RGBA, GL_UNSIGNED_BYTE, (unsigned char *)image->pixels);
	if (filter == GL_LINEAR) {
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	} else {
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_LINEAR);
	}
	
	/* libera la memoria che non ci serve piu' */
	SDL_UnlockSurface(image);
	SDL_FreeSurface(image);
	SDL_FreeSurface(bmp);
	return newTex;
}

// environment spherical map (prove)
void EnableTexEnvMap(void)
{
//	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
	/* abilita la mappatura sferica della texture */
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);
}

void DisableTexEnvMap(void)
{
//	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
}

// costruisci la texture lightmap per un poligono,
// considerando tutte le sorgenti di luce e gli altri
// poligoni per le ombre portate (TODO...)
// Per ora consideriamo solo la luce del sole (con relative ombre)
void BuildLightmap(Polygon3D *poly, ShadowPoly *sbuf, unsigned int scount)
{
	GLuint			tex = 0;
	Vector3D		ldir = gGlobalLight.dir, n = poly->normal;
	unsigned char	*img = NULL;
	unsigned int	res = 0, x, y, i;
	float			dot = DotProduct(ldir, n);
	float			nx = fabsf(n.x), ny = fabsf(n.y), nz = fabsf(n.z);
	float			minX, maxX, minY, maxY, minZ, maxZ, deltaX, deltaY, deltaZ;
	enum {XY = 0, XZ, YZ}	pplane;

	if (scount == 0)
		res = 1;	/* niente ombre, basta 1 pixel */
	else
		res = LIGHTMAP_RES;
	img = (unsigned char *)malloc(res * res * 3);
	if (img == NULL) {
		fprintf(stderr, "BuildLightmap: malloc(): %s\n", strerror(errno));
		return;
	}

	/* prima di tutto dobbiamo generare delle coordinate
	   u,v per ogni vertice del poligono, per mappare
	   la lightmap in maniera un po' "sensata" */
	
	// Per generare la lightmap, proiettiamo per semplicita'
	// il poligono su uno dei tre piani cartesiani, in particolare
	// quello in cui la proiezione e' piu' grande
	// (per questioni di precisione)
	if (nx > ny && nx > nz) {
		/* piano YZ */
		minY = maxY = poly->vertices[0].y;
		minZ = maxZ = poly->vertices[0].z;
		for (i = 0; i < poly->numVertices; i++) {
			Vertex3D	*v = &poly->vertices[i];

			if (v->y < minY)		minY = v->y;
			else if (v->y > maxY)	maxY = v->y;
			if (v->z < minZ)		minZ = v->z;
			else if (v->z > maxZ)	maxZ = v->z;
		}
		deltaY = maxY - minY;
		deltaZ = maxZ - minZ;
		for (i = 0; i < poly->numVertices; i++) {
			poly->vertices[i].lm_v = (poly->vertices[i].y - minY) / deltaY;
			poly->vertices[i].lm_u = (poly->vertices[i].z - minZ) / deltaZ;
		}
		pplane = YZ;
	} else if ((nx > ny && nx <= nz) ||
				(nx <= ny && ny <= nz)) {
		/* XY */
		minX = maxX = poly->vertices[0].x;
		minY = maxY = poly->vertices[0].y;
		for (i = 0; i < poly->numVertices; i++) {
			Vertex3D	*v = &poly->vertices[i];
			
			if (v->x < minX)		minX = v->x;
			else if (v->x > maxX)	maxX = v->x;
			if (v->y < minY)		minY = v->y;
			else if (v->y > maxY)	maxY = v->y;
		}
		deltaX = maxX - minX;
		deltaY = maxY - minY;
		for (i = 0; i < poly->numVertices; i++) {
			poly->vertices[i].lm_u = (poly->vertices[i].x - minX) / deltaX;
			poly->vertices[i].lm_v = (poly->vertices[i].y - minY) / deltaY;
		}
		pplane = XY;
	} else if (nx <= ny && ny > nz) {
		/* XZ */
		/* calcola il rettangolo di piano che contiene
		   esattamente la proiezione del poligono su XZ;
		   questo rettangolo diventera' la lightmap */
		minX = maxX = poly->vertices[0].x;
		minZ = maxZ = poly->vertices[0].z;
		for (i = 0; i < poly->numVertices; i++) {
			Vertex3D	*v = &poly->vertices[i];
			
			if (v->x < minX)		minX = v->x;
			else if (v->x > maxX)	maxX = v->x;
			if (v->z < minZ)		minZ = v->z;
			else if (v->z > maxZ)	maxZ = v->z;
		}
		deltaX = maxX - minX;
		deltaZ = maxZ - minZ;
		/* adesso ripassa ogni vertice, e calcola
		   le sue coordinate u,v per la lightmap
		   riproiettando e interpolando la corrispondente
		   proiezione relativamente al rettangolo trovato sopra */
		for (i = 0; i < poly->numVertices; i++) {
			poly->vertices[i].lm_u = (poly->vertices[i].x - minX) / deltaX;
			poly->vertices[i].lm_v = (poly->vertices[i].z - minZ) / deltaZ;
		}
		pplane = XZ;
	}
	
	/* illumina la lightmap in base alla normale e alla direzione della luce */
	for (y = 0; y < res; y++) {
		unsigned int	scanline = 3 * y * res;
		
		for (x = 0; x < res; x++) {
			float	r, g, b, brt;

			/* questa e' la nostra equazione del modello di illuminazione */
			brt = (dot > 0.0) ? 0.0 : (-dot);
			r = gGlobalLight.ambient[0] + brt * gGlobalLight.diffuse[0];
			if (r > 1.0)	r = 1.0;
			g = gGlobalLight.ambient[1] + brt * gGlobalLight.diffuse[1];
			if (g > 1.0)	g = 1.0;
			b = gGlobalLight.ambient[2] + brt * gGlobalLight.diffuse[2];
			if (b > 1.0)	b = 1.0;
			
			img[scanline + 3*x] = r * 255.0;
			img[scanline + 3*x + 1] = g * 255.0;
			img[scanline + 3*x + 2] = b * 255.0;
		}
	}
	
	/* disegna ogni shadow_poly nel buffer della lightmap... */
	for (i = 0; i < scount; i++) {
		Point2D			svl[kMaxVertices];	// vertici ombra in spazio lightmap
		Point3D			*sv = sbuf[i].vertices;
		unsigned int	svcount = sbuf[i].numVertices, j;
		int				y, minxs[res], maxxs[res];
		
		/* trasforma i vertici di questo ShadowPoly
		   nello spazio della lightmap */
		switch (pplane) {
			case XY:	// azzera la Z
				for (j = 0; j < svcount; j++) {
					svl[j].x = (sv[j].x - minX) * res / deltaX;
					svl[j].y = (sv[j].y - minY) * res / deltaY;
				}
				break;
			case XZ:	// azzera la Y
				for (j = 0; j < svcount; j++) {
					svl[j].x = (sv[j].x - minX) * res / deltaX;
					svl[j].y = (sv[j].z - minZ) * res / deltaZ;
				}
				break;
			case YZ:	// azzera la X
				for (j = 0; j < svcount; j++) {
					svl[j].y = (sv[j].y - minY) * res / deltaY;
					svl[j].x = (sv[j].z - minZ) * res / deltaZ;
				}
				break;
		}
		/* disegna il poligono cosi' ottenuto nel buffer della lightmap
		   (banale algoritmo a scanline per disegnare un poligono) */
		for (j = 0; j < res; j++) {
			minxs[j] = res;
			maxxs[j] = 0;
		}
		for (j = 0; j < svcount; j++) {
			int	x0, x1, y0, y1, x, y;

			x0 = svl[j].x;	y0 = svl[j].y;
			if (j == svcount-1) {
				x1 = svl[0].x;	y1 = svl[0].y;
			} else {
				x1 = svl[j+1].x;	y1 = svl[j+1].y;
			}
			if (y1 < y0) {
				int	temp = y1;

				y1 = y0;	y0 = temp;
				temp = x1;
				x1 = x0;	x0 = temp;
			}
			if (y0 >= res)	y0 = res-1;
			if (y1 >= res)	y1 = res-1;
			if (y1 != y0) {
				for (y = y0; y <= y1; y++) {
					x = (y - y0) * (x1 - x0) / (y1 - y0) + x0;
					if (x < 0)				x = 0;
					if (x >= res)	x = res-1;

					if (x < minxs[y])	minxs[y] = x;
					if (x > maxxs[y])	maxxs[y] = x;
				}
			} else {
				/* lato orizzontale */
				if (x0 < minxs[y0])	minxs[y0] = x0;
				if (x0 > maxxs[y0])	maxxs[y0] = x0;
				if (x1 < minxs[y0])	minxs[y0] = x1;
				if (x1 > maxxs[y0]) maxxs[y0] = x1;
			}
		}
		for (y = 0; y < res; y++) {
			unsigned int	scanline = 3 * y * res;
			
			if (minxs[y] <= maxxs[y]) {
				for (x = minxs[y]; x <= maxxs[y]; x++) {
					if (x >= 0 && x < res) {
						img[scanline + 3*x] = gGlobalLight.ambient[0] * 255.0;
						img[scanline + 3*x + 1] = gGlobalLight.ambient[1] * 255.0;
						img[scanline + 3*x + 2] = gGlobalLight.ambient[2] * 255.0;;
					}
				}
			}
		}
	}
	
	/* e finalmente generiamo la texture OpenGL a partire dal buffer img */
	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D, tex);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, 0, 3, res, res, 0, GL_RGB, GL_UNSIGNED_BYTE, img);
	poly->lmap = tex;

	free(img);
}

